# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import re

from packaging.version import Version


def parse_rss_version(rss_version: str) -> tuple[str, Version]:
    """Takes an RSS version string and returns a corresponding Version

    The RSS version string is on a 'git describe'-like format:

        RL-vA.B.C<-rcD><-E-gF>

    where

        RL: release line
        A:  major
        B:  minor
        C:  micro
        D:  release candidate,
        E:  additional commits since tag, F: commit SHA

    The concept of 'additional commits since tag' (E) doesn't have an
    equivalent in packaging.version.Version. Instead, when E is present,
    the smallest version part (D if present, otherwise C) is bumped and
    E is presented as a development prerelease.

    The commit SHA (F), if present, is translated to a 'local segment'.

    Examples:

    >>> parse_rss_version("a121-v1.2.3")
    ('a121', <Version('1.2.3')>)

    >>> parse_rss_version("a111-v2.10.3-rc4")
    ('a111', <Version('2.10.3rc4')>)

    >>> parse_rss_version("a121-v1.2.3-123-g0e03503be1")
    ('a121', <Version('1.2.4.dev123+g0e03503be1')>)

    >>> parse_rss_version("a121-v1.2.3-rc4-123-g0e03503be1")
    ('a121', <Version('1.2.3rc5.dev123+g0e03503be1')>)

    Read more: https://packaging.pypa.io/en/latest/version.html
    """

    pattern = (
        r"(?P<release_line>a\d{3})-v(?P<major>\d+)\.(?P<minor>\d+)\.(?P<micro>\d+)"
        r"(?:-(?P<pre_phase>rc)(?P<pre_number>\d+))?"
        r"(?:-(?P<dev_number>\d+)-(?P<dev_commit>g\w+))?"
        r"(?:-(dirty))?"
        r".*"
    )
    match = re.fullmatch(pattern, rss_version)
    if not match:
        msg = "Not a valid RSS version"
        raise ValueError(msg)

    groups = match.groupdict()

    is_prerelease = groups["pre_number"] is not None
    is_devrelease = groups["dev_number"] is not None

    release_segment = ""
    pre_segment = ""
    dev_segment = ""

    if is_devrelease:
        dev_segment = f".dev{groups['dev_number']}+{groups['dev_commit']}"

        if is_prerelease:
            groups["pre_number"] = int(groups["pre_number"]) + 1
        else:
            groups["micro"] = int(groups["micro"]) + 1

    if is_prerelease:
        pre_segment = f"{groups['pre_phase']}{groups['pre_number']}"

    release_segment = f"{groups['major']}.{groups['minor']}.{groups['micro']}"

    version = release_segment + pre_segment + dev_segment
    return (groups["release_line"], Version(version))



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Optional, Tuple, Union

import attrs
import h5py
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121.algo import (
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo.presence import (
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
)

from ._example_app import DetectionState, ExampleApp, ExampleAppConfig, ExampleAppResult


class AppMode(enum.Enum):
    """Indicating the current operating mode of the application."""

    PRESENCE = enum.auto()
    """Detecting presence."""

    HANDMOTION = enum.auto()
    """Detecting hand motion."""


@attributes_doc
@attrs.mutable(kw_only=True)
class ModeHandlerConfig(AlgoConfigBase):
    example_app_config: ExampleAppConfig = attrs.field(factory=ExampleAppConfig)
    """Hand motion detection configuration."""

    presence_config: DetectorConfig = attrs.field(factory=DetectorConfig)
    """Presence detector configuration."""

    hand_detection_timeout: Optional[float] = attrs.field(default=5.0)
    """Duration without detection before returning back to low power mode (s)."""

    use_presence_detection: Optional[bool] = attrs.field(default=True)
    """If true, use presence detector to wake up hand motion detection. """

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        return validation_results

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> ModeHandlerConfig:
        presence_dict = d["presence_config"]
        if presence_dict is not None:
            d["presence_config"] = DetectorConfig.from_dict(presence_dict)

        example_app_dict = d["example_app_config"]
        if example_app_dict is not None:
            d["example_app_config"] = ExampleAppConfig.from_dict(example_app_dict)
        return cls(**d)


opser.register_json_presentable(ModeHandlerConfig)


@attrs.frozen(kw_only=True)
class ModeHandlerResult:
    app_mode: AppMode
    """Application mode."""

    detection_state: Optional[DetectionState] = attrs.field(default=None)
    """Detection state."""

    presence_result: Optional[DetectorResult] = attrs.field(default=None)
    """Presence detector result."""

    example_app_result: Optional[ExampleAppResult] = attrs.field(default=None)
    """Hand motion example app result."""


class ModeHandler(Controller[ModeHandlerConfig, ModeHandlerResult]):
    """
    Application handling the switching between the presence detector and the hand motion
    application.

    :param client: Client
    :param sensor_id: Sensor id
    :param mode_handler_config: Mode handler configuration
    """

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        mode_handler_config: ModeHandlerConfig,
    ) -> None:
        self.client = client
        self.sensor_id = sensor_id
        self.config = mode_handler_config

        self.example_app_config = mode_handler_config.example_app_config
        self.presence_config = mode_handler_config.presence_config

        # As presence detector is used to lower power consumption, it is assumed that the frame
        # rate is low and can be explicitly set here, rather than estimated in the presence
        # application, introducing latency.
        self.presence_detector_context = DetectorContext(
            estimated_frame_rate=mode_handler_config.presence_config.frame_rate
        )

        assert mode_handler_config.hand_detection_timeout is not None
        # Convert timeout from seconds to frames.
        self.hand_motion_timeout_duration = int(
            round(mode_handler_config.hand_detection_timeout * self.example_app_config.frame_rate)
        )
        self.use_presence_detection = mode_handler_config.use_presence_detection

        if self.use_presence_detection:
            self.app_mode = AppMode.PRESENCE
        else:
            self.app_mode = AppMode.HANDMOTION
        self.hand_motion_timer = 0
        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        """Start application."""
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        if _algo_group is None and isinstance(recorder, a121.H5Recorder):
            _algo_group = recorder.require_algo_group("hand_motion")

        if recorder is not None:
            self.client.attach_recorder(recorder)

        if _algo_group is not None:
            _record_algo_data(_algo_group, self.sensor_id, self.config)

        if self.use_presence_detection:
            self.presence_detector = Detector(
                client=self.client,
                sensor_id=self.sensor_id,
                detector_config=self.presence_config,
                detector_context=self.presence_detector_context,
            )
            self.presence_detector.start()
        else:
            self.hand_motion_app = ExampleApp(
                client=self.client,
                sensor_id=self.sensor_id,
                example_app_config=self.example_app_config,
            )
            self.hand_motion_app.start()

        self.started = True

    def stop(self) -> Any:
        """Stop application."""
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        if self.use_presence_detection and self.presence_detector.started:
            recorder_result = self.presence_detector.stop()
        elif self.hand_motion_app.started:
            recorder_result = self.hand_motion_app.stop()
        else:
            msg = "No active session to stop."
            raise RuntimeError(msg)

        self.started = False

        return recorder_result

    def get_next(self) -> ModeHandlerResult:
        """Get next result."""
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        if self.app_mode == AppMode.PRESENCE:
            presence_result = self.presence_detector.get_next()
            result = self._determine_mode_swap(app_result=presence_result)
        elif self.app_mode == AppMode.HANDMOTION:
            hand_motion_result = self.hand_motion_app.get_next()
            result = self._determine_mode_swap(app_result=hand_motion_result)
        else:
            msg = "Invalid app"
            raise RuntimeError(msg)

        return result

    def _determine_mode_swap(
        self, app_result: Union[DetectorResult, ExampleAppResult]
    ) -> ModeHandlerResult:
        """Determine if mode should be swapped, based on result."""
        current_app_mode = self.app_mode

        # Determine if mode should be swapped
        if self.app_mode == AppMode.PRESENCE:
            assert isinstance(app_result, DetectorResult)
            if app_result.presence_detected:
                self.app_mode = AppMode.HANDMOTION
            result = ModeHandlerResult(app_mode=current_app_mode, presence_result=app_result)
        elif self.app_mode == AppMode.HANDMOTION:
            assert isinstance(app_result, ExampleAppResult)
            if self.use_presence_detection:
                if app_result.detection_state is not DetectionState.NO_DETECTION:
                    # detection -> reset timer
                    self.hand_motion_timer = 0
                elif self.hand_motion_timeout_duration < self.hand_motion_timer:
                    # timer has expired -> switch mode
                    self.app_mode = AppMode.PRESENCE
                else:
                    self.hand_motion_timer += 1
            result = ModeHandlerResult(
                app_mode=current_app_mode,
                detection_state=app_result.detection_state,
                example_app_result=app_result,
            )
        else:
            msg = "Invalid app mode"
            raise RuntimeError(msg)

        if self.app_mode != current_app_mode:
            self._swap_mode()

        return result

    def _swap_mode(self) -> None:
        """Swaps mode by stopping current application and create and start a new application."""
        if self.app_mode == AppMode.PRESENCE:
            self.hand_motion_app.stop_detector()
            self.presence_detector = Detector(
                client=self.client,
                sensor_id=self.sensor_id,
                detector_config=self.presence_config,
                detector_context=self.presence_detector_context,
            )
            self.presence_detector.start()
        elif self.app_mode == AppMode.HANDMOTION:
            self.presence_detector.stop_detector()
            self.hand_motion_app = ExampleApp(
                client=self.client,
                sensor_id=self.sensor_id,
                example_app_config=self.example_app_config,
            )
            self.hand_motion_app.start()
            self.hand_motion_timer = 0
        else:
            msg = "Invalid app"
            raise RuntimeError(msg)


def get_default_config() -> ModeHandlerConfig:
    # Create presence config with low power and high responsiveness.
    presence_config = DetectorConfig()
    presence_config.frame_rate = 2
    presence_config.start_m = 0.7
    presence_config.end_m = 0.7
    presence_config.sweeps_per_frame = 32
    presence_config.hwaas = 4
    presence_config.inter_frame_deviation_time_const = 0.01
    presence_config.inter_output_time_const = 0.01
    presence_config.intra_frame_time_const = 0.01
    presence_config.intra_output_time_const = 0.01

    return ModeHandlerConfig(presence_config=presence_config)


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ModeHandlerConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )

    handle_config_group = algo_group.create_group("mode_handler_config")
    opser.serialize(config, handle_config_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ModeHandlerConfig]:
    sensor_id = int(algo_group["sensor_id"][()])
    config = opser.deserialize(algo_group["mode_handler_config"], ModeHandlerConfig)
    return sensor_id, config



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QGridLayout, QLabel, QToolButton, QWidget


class CollapsibleWidget(QWidget):
    _ICON_SIZE = 28

    def __init__(self, label: str, widget: QWidget, parent: t.Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.widget = widget
        self.widget.setVisible(False)

        self.arrow_button = QToolButton()
        self.arrow_button.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.arrow_button.setText(label)
        self.arrow_button.setArrowType(QtCore.Qt.ArrowType.RightArrow)
        self.arrow_button.setCheckable(True)
        self.arrow_button.setChecked(False)
        self.arrow_button.toggled.connect(self._on_toggle)

        self.icon_label = QLabel()
        self.set_icon(None)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.arrow_button, 0, 0)
        layout.addWidget(self.icon_label, 0, 1)
        layout.addWidget(self.widget, 1, 0, 1, -1)

        self.setLayout(layout)

    def set_collapsed(self, collapsed: bool) -> None:
        self.widget.setVisible(not collapsed)
        self.arrow_button.setArrowType(
            QtCore.Qt.ArrowType.RightArrow if collapsed else QtCore.Qt.ArrowType.DownArrow
        )

    def set_icon(self, icon: t.Optional[QtGui.QIcon]) -> None:
        if icon is not None:
            self.icon_label.setVisible(True)
            self.icon_label.setPixmap(icon.pixmap(self._ICON_SIZE))
        else:
            self.icon_label.setVisible(False)

    def _on_toggle(self, checked: bool) -> None:
        self.set_collapsed(not checked)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import enum
import inspect
import json
import logging
from copy import copy
from types import SimpleNamespace

import attr
import numpy as np


log = logging.getLogger(__name__)


class Category(enum.Enum):
    BASIC = enum.auto()
    ADVANCED = enum.auto()


class Severity(enum.IntEnum):
    ERROR = enum.auto()
    WARNING = enum.auto()
    INFO = enum.auto()


@attr.s
class Alert:
    param = attr.ib()
    msg = attr.ib()


class Error(Alert):
    severity = Severity.ERROR


class Warning(Alert):
    severity = Severity.WARNING


class Info(Alert):
    severity = Severity.INFO


class Parameter:
    def __init__(self, **kwargs):
        self.label = kwargs.pop("label")
        self.is_live_updateable = kwargs.pop("updateable", False)
        self.does_dump = kwargs.pop("does_dump", False)
        self.category = kwargs.pop("category", Category.BASIC)
        self.order = kwargs.pop("order", -1)
        self.help = kwargs.pop("help", None)
        self.visible = kwargs.pop("visible", True)
        self.is_dummy = kwargs.pop("is_dummy", False)
        self.enabled = kwargs.pop("enabled", True)

        self._pidget_class = kwargs.pop("_pidget_class", None)

        # don't care if unused
        kwargs.pop("default_value", None)

        if kwargs:
            a_key = next(iter(kwargs.keys()))
            msg = "Got unexpected keyword argument ({})".format(a_key)
            raise TypeError(msg)

        if not self.visible:
            self._pidget_class = None
        elif self._pidget_class is None:
            self.visible = False

        doc = self.generate_doc()
        if doc:
            self.__doc__ = doc.strip()

        self.help_obj = SimpleNamespace()
        self.help_obj.__doc__ = None if self.help is None else inspect.cleandoc(self.help)

        self._attr_name = None

    @property
    def pidget_attr_key(self):
        return "__pidget_" + self._attr_name

    def update_pidget(self, obj, alerts=None):
        pidget = self.get_pidget(obj)
        if pidget is not None:
            pidget.update(alerts)

    def get_pidget(self, obj):
        return obj.__dict__.get(self.pidget_attr_key)

    def create_pidget(self, obj):
        from . import qtpidgets

        if self._pidget_class is None:
            return None

        key = self.pidget_attr_key

        if key not in obj.__dict__:
            pidget_class = getattr(qtpidgets, self._pidget_class)
            obj.__dict__[key] = pidget_class(self, obj)

        return obj.__dict__[key]

    def pidget_event_handler(self, *args, **kwargs):
        pass

    def dump(self):
        pass

    def load(self):
        pass

    def generate_doc(self):
        if self.help is None:
            return None

        return inspect.cleandoc(self.help)


class ConstantParameter(Parameter):
    def __init__(self, **kwargs):
        self.value = kwargs.pop("value")
        super().__init__(**kwargs)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        return self.value

    def __set__(self, obj, value):
        msg = "Unsettable parameter"
        raise AttributeError(msg)


class ValueParameter(Parameter):
    def __init__(self, **kwargs):
        self.default_value = kwargs.pop("default_value")
        self.is_optional = kwargs.pop("optional", False)

        if self.is_optional:
            self.optional_label = kwargs.pop("optional_label", "Set")

            if self.default_value is None:
                self.optional_default_set_value = kwargs.pop("optional_default_set_value")
            else:
                self.optional_default_set_value = self.default_value

        kwargs.setdefault("does_dump", True)

        super().__init__(**kwargs)

        self.default_value = self.sanitize(self.default_value)

    @property
    def value_attr_key(self):
        return "__value_" + self._attr_name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        return copy(obj.__dict__.get(self.value_attr_key, self.default_value))

    def __set__(self, obj, value):
        value = self.sanitize(value)
        obj.__dict__[self.value_attr_key] = value

    def __delete__(self, obj):
        obj.__dict__.pop(self.value_attr_key, None)

    def pidget_event_handler(self, obj, val):
        self.__set__(obj, val)  # might raise an exception

        obj._parameter_event_handler()

    def sanitize(self, value):
        if not (self.is_optional and value is None):
            value = self._sanitize(value)

        return value

    def _sanitize(self, value):
        return value

    def dump(self, obj):
        return self.__get__(obj)

    def load(self, obj, value):
        self.__set__(obj, value)

    def generate_doc(self):
        s = ""

        s_help = super().generate_doc()
        if s_help:
            s += s_help
            s += "\n\n"

        type_str = getattr(self, "type_str", None)
        if type_str:
            s += "| Type: {}\n".format(type_str)

        unit = getattr(self, "unit", None)
        if unit:
            s += "| Unit: {}\n".format(unit)

        s += "| Default value: {}\n".format(self.default_value)

        return s


class BoolParameter(ValueParameter):
    type_str = "bool"

    def __init__(self, **kwargs):
        kwargs.setdefault("_pidget_class", "BoolCheckboxPidget")

        super().__init__(**kwargs)

    def _sanitize(self, value):
        return bool(value)


class EnumParameter(ValueParameter):
    def __init__(self, **kwargs):
        self.enum = kwargs.pop("enum")

        kwargs.setdefault("_pidget_class", "ComboBoxPidget")

        super().__init__(**kwargs)

    def _sanitize(self, value):
        if isinstance(value, self.enum):
            return value

        if not isinstance(value, str):
            msg = "Must be a {} or str (member name)".format(self.enum.__name__)
            raise ValueError(msg)

        try:
            return self.enum[value.upper()]
        except KeyError as e:
            msg = "{} is not a valid {} member name".format(value, self.enum.__name__)
            raise ValueError(msg) from e

    def dump(self, obj):
        return self.__get__(obj).name

    def load(self, obj, value):
        value = self.enum[value]
        self.__set__(obj, value)


class NumberParameter(ValueParameter):
    def __init__(self, **kwargs):
        self.unit = kwargs.pop("unit", None)
        self.limits = kwargs.pop("limits", None)

        if self.limits is not None:
            assert isinstance(self.limits, (tuple, list))
            assert len(self.limits) == 2

        super().__init__(**kwargs)


class IntParameter(NumberParameter):
    type_str = "int"

    def __init__(self, **kwargs):
        self.step = kwargs.pop("step", 1)
        self.valid_values = kwargs.pop("valid_values", None)

        limits = kwargs.get("limits", None)
        if self.valid_values is not None and limits is not None:
            msg = "valid_values and limits can not both be set on the same parameter"
            raise ValueError(msg)

        if self.valid_values is not None:
            kwargs.setdefault("_pidget_class", "IntComboBoxPidget")
        else:
            kwargs.setdefault("_pidget_class", "IntSpinBoxPidget")

        super().__init__(**kwargs)

    def _sanitize(self, value):
        if isinstance(value, float) and not value.is_integer():
            msg = "Not an integer"
            raise ValueError(msg)

        try:
            value = int(value)
        except ValueError:
            msg = "Not a valid integer"
            raise ValueError(msg)

        if self.limits is not None:
            lower, upper = self.limits

            if lower is not None and value < lower:
                msg = "Given value is too low ({} < {})".format(value, lower)
                raise ValueError(msg)
            if upper is not None and value > upper:
                msg = "Given value is too high ({} > {})".format(value, upper)
                raise ValueError(msg)

        if self.valid_values is not None and value not in self.valid_values:
            msg = "Given value ({}) is not a valid value".format(value)
            raise ValueError(msg)

        return value


class FloatParameter(NumberParameter):
    type_str = "float"

    def __init__(self, **kwargs):
        self.decimals = kwargs.pop("decimals", 2)
        self.logscale = kwargs.pop("logscale", False)

        optional = kwargs.get("optional", False)

        limits = kwargs.get("limits", None)
        if limits is not None and not any([lim is None for lim in limits]) and not optional:
            kwargs.setdefault("_pidget_class", "FloatSpinBoxAndSliderPidget")
        else:
            kwargs.setdefault("_pidget_class", "FloatSpinBoxPidget")

        super().__init__(**kwargs)

        if self.logscale:
            assert self.limits is not None
            assert self.limits[1] > self.limits[0] > 0

    def _sanitize(self, value):
        try:
            value = float(value)
        except ValueError:
            msg = "Not a valid number"
            raise ValueError(msg)

        value = round(value, self.decimals)

        if self.limits is not None:
            lower, upper = self.limits

            if lower is not None and value < lower:
                msg = "Given value is too low ({} < {})".format(value, lower)
                raise ValueError(msg)
            if upper is not None and value > upper:
                msg = "Given value is too high ({} > {})".format(value, upper)
                raise ValueError(msg)

        return value


class FloatRangeParameter(FloatParameter):
    def __init__(self, **kwargs):
        kwargs.setdefault("_pidget_class", "FloatRangeSpinBoxesPidget")

        super().__init__(**kwargs)

    def _sanitize(self, arg):
        try:
            values = list(arg)
        except (ValueError, TypeError):
            msg = "Not a valid range"
            raise ValueError(msg)

        if len(values) != 2:
            msg = "Given range does not have two values"
            raise ValueError(msg)

        for i in range(2):
            values[i] = super()._sanitize(values[i])

        if values[0] > values[1]:
            msg = "Invalid range"
            raise ValueError(msg)

        return np.array(values)

    def dump(self, obj):
        return list(self.__get__(obj))


class SensorParameter(ValueParameter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _sanitize(self, arg):
        if isinstance(arg, int):
            arg = [arg]
        elif isinstance(arg, list) and all([isinstance(e, int) for e in arg]):
            arg = copy(arg)
        else:
            msg = "sensor(s) must be an int or a list of ints"
            raise ValueError(msg)

        return arg


class ClassParameter(Parameter):
    def __init__(self, **kwargs):
        self.objtype = kwargs.pop("objtype")

        super().__init__(**kwargs)

    @property
    def instance_attr_key(self):
        return "__instance_" + self._attr_name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        key = self.instance_attr_key

        if key not in obj.__dict__:
            obj.__dict__[key] = self.objtype(self, obj)

        return obj.__dict__[key]

    def __set__(self, obj, value):
        msg = "Unsettable parameter"
        raise AttributeError(msg)

    def __delete__(self, obj):
        obj.__dict__.pop(self.instance_attr_key, None)


def get_virtual_parameter_class(base_class):
    assert issubclass(base_class, ValueParameter)

    class VirtualParameter(base_class):
        def __init__(self, **kwargs):
            self.get_fun = kwargs.pop("get_fun")
            self.set_fun = kwargs.pop("set_fun", None)

            kwargs.setdefault("does_dump", False)

            kwargs.setdefault("default_value", None)

            super().__init__(**kwargs)

        def __get__(self, obj, objtype=None):
            if obj is None:
                return self

            return self.get_fun(obj)

        def __set__(self, obj, value):
            if self.set_fun is None:
                msg = "Unsettable parameter"
                raise AttributeError(msg)

            self.set_fun(obj, value)

        def __delete__(self, obj):
            pass

        def _sanitize(self, value):
            pass

        def generate_doc(self):
            return None

    return VirtualParameter


class ConfigMeta(type):
    def __new__(cls, name, bases, d):
        for key, val in d.items():
            if isinstance(val, Parameter):
                val._attr_name = key

        return super(ConfigMeta, cls).__new__(cls, name, bases, d)


class Config(metaclass=ConfigMeta):
    class State(enum.Enum):
        UNLOADED = enum.auto()
        LOADED = enum.auto()
        LOADED_READONLY = enum.auto()
        LIVE = enum.auto()

    VERSION = None

    _event_handlers = None
    __state = None

    def __init__(self):
        self._event_handlers = set()
        self.__state = Config.State.UNLOADED

    def __str__(self):
        d = {k: p.dump(self) for k, p in self._get_keys_and_params() if p.does_dump}
        s = self.__class__.__name__
        s += "".join(["\n  {:.<35} {}".format(a + " ", v) for (a, v) in d.items()])
        return s

    def _loads(self, s):
        d = json.loads(s)

        version = d.pop("VERSION", None)
        if version != self.VERSION:
            msg = "Configuration version mismatch"
            raise ValueError(msg)

        params = dict(self._get_keys_and_params())
        for k, v in d.items():
            if k not in params:
                log.warning("Skipping unknown parameter {} when loading dump".format(k))
                continue

            params[k].load(self, v)

        self._update_pidgets()

    def _dumps(self):
        d = {k: p.dump(self) for k, p in self._get_keys_and_params() if p.does_dump}

        if self.VERSION is not None:
            d["VERSION"] = self.VERSION

        return json.dumps(d)

    def _reset(self):
        params = self._get_params()
        for param in params:
            param.__delete__(self)
            param.update_pidget(self)

        self._parameter_event_handler()

    def _create_pidgets(self):
        params = self._get_params()
        pidgets = [param.create_pidget(self) for param in params]
        return pidgets

    def _update_pidgets(self, additional_alerts=[]):
        alerts = self.check()
        if alerts is None:
            alerts = []
        alerts.extend(additional_alerts)

        for key, param in self._get_keys_and_params():
            param_alerts = [a for a in alerts if a.param in [key, param]]
            param.update_pidget(self, param_alerts)

        return alerts

    def _parameter_event_handler(self):
        for event_handler in self._event_handlers:
            event_handler(self)

    def _get_keys_and_params(self):
        keys = dir(self)
        attrs = [getattr(type(self), key, None) for key in keys]
        z = [(k, a) for k, a in zip(keys, attrs) if isinstance(a, Parameter)]
        return sorted(z, key=lambda t: t[1].order)

    def _get_params(self):
        return [a for k, a in self._get_keys_and_params()]

    @property
    def _state(self):
        return self.__state

    @_state.setter
    def _state(self, state):
        state_changed = self.__state != state
        self.__state = state
        if state_changed:
            self._update_pidgets()

    def check(self):
        return []

    def __setattr__(self, name, value):
        if hasattr(self, name):
            object.__setattr__(self, name, value)
        else:
            fmt = "'{}' object has no attribute '{}'"
            raise AttributeError(fmt.format(self.__class__.__name__, name))


class SensorConfig(Config):
    pass


class ProcessingConfig(Config):
    pass



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from ._detector import Detector, DetectorConfig, DetectorMetadata, DetectorResult
from ._processors import Processor, ProcessorConfig, ProcessorResult



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import enum
import typing as t

from PySide6.QtCore import QPoint, QRect
from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QFrame, QGridLayout, QLabel, QLayout, QVBoxLayout, QWidget


_T = t.TypeVar("_T", bound=QLayout)


class GroupBox(QWidget, t.Generic[_T]):
    _HORIZONTAL_HEADER_MARGIN = 7
    _EXTRA_MARGIN_TO_AVOID_CLIPPING = 2

    class _Position(enum.Enum):
        LEFT = enum.auto()
        RIGHT = enum.auto()

    def __init__(
        self,
        left_header: t.Union[QWidget, str],
        layout_type: t.Type[_T],
        right_header: t.Optional[QWidget] = None,
        *,
        min_top_padding: int = 0,
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self._own_layout = QVBoxLayout()
        self.setLayout(self._own_layout)

        self._frame_layout = layout_type()
        self._frame = QFrame()
        self._frame.setFrameShape(QFrame.Shape.Box)
        self._frame.setLayout(self._frame_layout)
        self._own_layout.addWidget(self._frame)

        # Header widgets are not added to the layout as their positions are handled manually
        self._header_frames = {}
        self._min_top_padding = min_top_padding

        if isinstance(left_header, QWidget):
            self._header_frames[self._Position.LEFT] = self._wrap_in_frame(left_header)
        elif isinstance(left_header, str) and left_header != "":
            self._header_frames[self._Position.LEFT] = self._wrap_in_frame(
                QLabel(f"<b>{left_header}</b>"),
            )

        if right_header is not None:
            self._header_frames[self._Position.RIGHT] = self._wrap_in_frame(right_header)

    @classmethod
    def vertical(
        cls,
        left_header: t.Union[QWidget, str],
        right_header: t.Optional[QWidget] = None,
        *,
        min_top_padding: int = 0,
        parent: t.Optional[QWidget] = None,
    ) -> GroupBox[QVBoxLayout]:
        return cls(
            left_header,
            QVBoxLayout,  # type: ignore[return-value,arg-type]
            right_header=right_header,
            min_top_padding=min_top_padding,
            parent=parent,
        )

    @classmethod
    def grid(
        cls,
        left_header: t.Union[QWidget, str],
        right_header: t.Optional[QWidget] = None,
        *,
        min_top_padding: int = 0,
        parent: t.Optional[QWidget] = None,
    ) -> GroupBox[QGridLayout]:
        return cls(
            left_header,
            QGridLayout,  # type: ignore[return-value,arg-type]
            right_header=right_header,
            min_top_padding=min_top_padding,
            parent=parent,
        )

    def layout(self) -> _T:
        """Returns the layout of the internal QFrame"""
        return self._frame.layout()  # type: ignore[return-value]

    def _wrap_in_frame(self, widget: QWidget) -> QFrame:
        """
        Header widgets are wrapped in a frame to
        make sure the header widgets have an opaque background
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)

        frame = QFrame(parent=self)
        frame.setObjectName("GroupBoxQFrameHeader")
        frame.setStyleSheet("#GroupBoxQFrameHeader { border: none; padding: 1px; }")
        frame.setFrameShape(QFrame.Shape.NoFrame)
        frame.setLayout(layout)

        return frame

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Handles positioning of header frame widgets manually"""
        super().resizeEvent(event)

        top_margin = max(
            [
                self._min_top_padding,
                *(frame.rect().center().y() for frame in self._header_frames.values()),
            ]
        )

        self._own_layout.setContentsMargins(0, top_margin, 0, 0)

        top_padding = top_margin
        frame_margins = self._frame_layout.contentsMargins()
        frame_margins.setTop(top_padding + self._EXTRA_MARGIN_TO_AVOID_CLIPPING)
        self._frame_layout.setContentsMargins(frame_margins)

        self._own_layout.activate()

        for position, frame in self._header_frames.items():
            new_rect = QRect(QPoint(), frame.sizeHint())
            new_rect.moveCenter(self._frame.pos())

            if position == self._Position.LEFT:
                new_rect.moveLeft(self._HORIZONTAL_HEADER_MARGIN)
            elif position == self._Position.RIGHT:
                new_rect.moveRight(self.width() - self._HORIZONTAL_HEADER_MARGIN)
            else:
                assert False, f"Unexpected HeaderWidgetPosition {position}"

            frame.setGeometry(new_rect)



# Copyright (c) Acconeer AB, 2025
# All rights reserved
from __future__ import annotations

import importlib.resources
import typing as t
from dataclasses import dataclass
from enum import Enum, IntEnum, auto

import requests

from PySide6.QtCore import QObject, QRunnable, QThreadPool, QTimerEvent, Signal
from PySide6.QtGui import QDesktopServices, QPixmap
from PySide6.QtWidgets import (
    QButtonGroup,
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QFormLayout,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QSizePolicy,
    QTextEdit,
    QToolButton,
    QVBoxLayout,
    QWidget,
    QWizard,
    QWizardPage,
)

from acconeer.exptool._core.communication.comm_devices import CommDevice
from acconeer.exptool.app.new.app import ConnectionState
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.storage import get_temp_dir
from acconeer.exptool.app.new.ui import icons
from acconeer.exptool.flash import (
    DevLicense,
    clear_cookies,
    download,
    flash_image,
    get_boot_description,
    get_content,
    get_cookies,
    get_flash_download_name,
    get_flash_known_devices,
    get_post_dfu_description,
    is_fw_downloadable,
    login,
    save_cookies,
)
from acconeer.exptool.flash._dev_license import DEV_LICENSE_DEFAULT_HEADER
from acconeer.exptool.flash._products import EVK_TO_PRODUCT_MAP

from . import resources
from .dialogs import UserMessageDialog


DEVSITE_REGISTER_URL = "https://developer.acconeer.com/register/"


class WizardField(str, Enum):
    DOWNLOAD_SELECTED = "download_selected"
    MODULE = "module"

    EMAIL = "email"
    PASSWORD = "password"
    REMEMBER_ME = "remember_me"
    LOGIN_FEEDBACK = "login_feedback"
    LOGIN_FEEDBACK_VISIBLE = "login_feedback_visible"
    LOGIN_FEEDBACK_STYLESHEET = "login_feedback_stylesheet"

    BIN_PATH = "bin_path"

    def required(self) -> str:
        return self + "*"


class FeedbackStyleSheet(str, Enum):
    ERROR = f"background-color: {icons.ERROR_RED}; color: white"
    NICE = f"background-color: {icons.BUTTON_ICON_COLOR}; color: white"


class ModuleSelectionPage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.setTitle("Select module to flash")

        with importlib.resources.path(resources, "xe125.jpg") as path:
            assert path.exists()
            self.xe125_pixmap = QPixmap(path)

        with importlib.resources.path(resources, "xe121.jpg") as path:
            assert path.exists()
            self.xe121_pixmap = QPixmap(path)

        with importlib.resources.path(resources, "xm126.jpg") as path:
            assert path.exists()
            self.xm126_pixmap = QPixmap(path)

        self.picture_label = QLabel("Picture")

        combo_box = QComboBox()
        combo_box.currentTextChanged.connect(self.switch_to_matching_pixmap)

        for supported_module_name in get_flash_known_devices():
            combo_box.addItem(supported_module_name)

        self.registerField(WizardField.MODULE, combo_box, "currentText", "currentTextChanged")

        layout = QFormLayout()
        layout.addRow("Module", combo_box)
        layout.addRow(self.picture_label)
        self.setLayout(layout)

    def switch_to_matching_pixmap(self, combo_box_text: str) -> None:
        if combo_box_text == "XC120":
            self.picture_label.setPixmap(self.xe121_pixmap)
        elif combo_box_text == "XM125":
            self.picture_label.setPixmap(self.xe125_pixmap)
        elif combo_box_text == "XM126":
            self.picture_label.setPixmap(self.xm126_pixmap)
        else:
            pass


class DeviceSelectionPage(QWizardPage):
    _TITLE_FMT = "Select a connected {} to flash"

    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self.app_model = app_model
        self.setTitle(self._TITLE_FMT.format("device"))

        self.device_cb = QComboBox()
        self.device_cb.currentTextChanged.connect(self.update_wizards_comm_device)
        self.device_cb.currentTextChanged.connect(lambda _: self.completeChanged.emit())

        refresh_button = QPushButton()
        refresh_button.setIcon(icons.REFRESH())
        refresh_button.clicked.connect(self._repopulate_combobox)

        self.show_all_devices_checkbox = QCheckBox("Show all devices")
        self.show_all_devices_checkbox.setChecked(False)
        self.show_all_devices_checkbox.toggled.connect(self._repopulate_combobox)

        self.feedback_lbl = QLabel("")
        self.feedback_lbl.setWordWrap(True)
        self.feedback_lbl.setStyleSheet(f"background-color: {icons.WARNING_YELLOW}; color: white")
        self.feedback_lbl.setVisible(False)

        cb_refresh_layout = QHBoxLayout()
        cb_refresh_layout.addWidget(self.device_cb, stretch=1)
        cb_refresh_layout.addWidget(refresh_button, stretch=0)

        layout = QFormLayout()
        layout.addRow("Selected device", cb_refresh_layout)
        layout.addRow(self.show_all_devices_checkbox)
        layout.addRow(self.feedback_lbl)
        layout.addRow(_Spacer())
        self.setLayout(layout)

    def isComplete(self) -> bool:
        w = self.wizard()
        assert isinstance(w, FlashWizard)
        return w.comm_device is not None

    def update_wizards_comm_device(self) -> None:
        w = self.wizard()
        assert isinstance(w, FlashWizard)
        w.set_comm_device(self.device_cb.currentData())

    def initializePage(self) -> None:
        self._repopulate_combobox()
        self.setTitle(self._TITLE_FMT.format(self.field(WizardField.MODULE)))

    def cleanupPage(self) -> None:
        self.show_all_devices_checkbox.setChecked(False)

    def is_currently_selected_module(self, comm_device: CommDevice) -> bool:
        if comm_device.name is None:
            return False

        evk_name = comm_device.name
        product_name = EVK_TO_PRODUCT_MAP[evk_name]
        return bool(self.field(WizardField.MODULE) == product_name)

    def _repopulate_combobox(self) -> None:
        self.device_cb.clear()
        comm_devices = [
            dev
            for dev in (
                self.app_model.available_usb_devices + self.app_model.available_serial_devices
            )
            if (self.is_currently_selected_module(dev) and dev.recognized)
            or self.show_all_devices_checkbox.isChecked()
        ]
        for device in comm_devices:
            self.device_cb.addItem(device.display_name(), device)

        if self.device_cb.count() == 0:
            self.feedback_lbl.setText(
                "<br>".join(
                    [
                        f"No <b>{self.field(WizardField.MODULE)}</b> can be recognized.",
                        "Is your module properly plugged in to this computer?",
                        "Try reinserting USB cable and the clicking the Refresh Button above."
                        "As a last resort, check <i>Show all devices</i> and look for your device there.",
                    ]
                )
            )
            self.feedback_lbl.setVisible(True)
        else:
            self.feedback_lbl.setVisible(False)


class SelectLocalFileOrDownloadFromDevsitePage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.setTitle("Application binary source")

        self._download_btn = QRadioButton("Automatically download from developer.acconeer.com")
        self._download_btn.setChecked(True)
        self.registerField(WizardField.DOWNLOAD_SELECTED, self._download_btn, "checked", "toggled")

        self._local_file_btn = QRadioButton("Local file")

        button_group = QButtonGroup(parent=self)
        button_group.addButton(self._download_btn)
        button_group.addButton(self._local_file_btn)

        layout = QVBoxLayout()
        layout.addWidget(self._download_btn)
        layout.addWidget(self._local_file_btn)
        self.setLayout(layout)

    def initializePage(self) -> None:
        if is_fw_downloadable(self.field(WizardField.MODULE)):
            self._download_btn.setEnabled(True)
            self._download_btn.setChecked(True)
        else:
            self._download_btn.setEnabled(False)
            self._local_file_btn.setChecked(True)


class BrowseLocalFilePage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        # The state management gets weird if going back to here from the FlashPage.
        # Making this a "commit page" makes it impossible to go back here
        # (and the user needs to cancel the wizard instead)
        self.setCommitPage(True)
        self.setButtonText(QWizard.WizardButton.CommitButton, "Next")

        self.path_label = QLineEdit()
        self.path_label.setReadOnly(True)
        self.path_label.setPlaceholderText("Path to local file")

        browse_btn = QPushButton("Browse")
        browse_btn.clicked.connect(self._select_local_file)

        layout = QHBoxLayout()
        layout.addWidget(self.path_label, stretch=1)
        layout.addWidget(browse_btn, stretch=0)
        self.setLayout(layout)

    def _select_local_file(self) -> None:
        filename, _ = QFileDialog.getOpenFileName(
            caption="caption", filter="Bin files (*.bin *.image)"
        )
        self.path_label.setText(filename)
        self.setField(WizardField.BIN_PATH, filename)
        self.completeChanged.emit()

    def isComplete(self) -> bool:
        return len(self.field(WizardField.BIN_PATH)) > 0

    def initializePage(self) -> None:
        self.setTitle(f"Browse for local binary to flash your {self.field('module')} with")


class LoginFormPage(QWizardPage):
    _LOGIN_BUTTON_DEFAULT_TEXT = "Log in"
    _LOGIN_BUTTON_PENDING_TEXT = "Logging in ..."

    _COOKIE_POLICY_LINK = "<a href='https://acconeer.com/cookie-policy-eu/'>Cookie Policy</a>"
    _PRIVACY_STATEMENT_LINK = (
        "<a href='https://developer.acconeer.com/privacy-policy/'>Privacy Statement</a>"
    )
    _REMEMBER_ME_TERMS_HTML = (
        "<i>"
        + (
            "<br>".join(
                [
                    '"Remember me" will save your credentials for next time.',
                    "By checking the box you agree to",
                    f"Acconeer's {_COOKIE_POLICY_LINK} and {_PRIVACY_STATEMENT_LINK}.",
                ]
            )
        )
        + "</i>"
    )

    def __init__(self) -> None:
        super().__init__()
        self.feedback_lbl = QLineEdit("Log in using your developer.accconeer.com account below")
        self.feedback_lbl.setReadOnly(True)
        self.feedback_lbl.setVisible(True)
        self.feedback_lbl.setStyleSheet(FeedbackStyleSheet.NICE)

        self.registerField(WizardField.LOGIN_FEEDBACK, self.feedback_lbl, "text")
        self.registerField(WizardField.LOGIN_FEEDBACK_VISIBLE, self.feedback_lbl, "visible")
        self.registerField(WizardField.LOGIN_FEEDBACK_STYLESHEET, self.feedback_lbl, "styleSheet")

        self.email_line_edit = QLineEdit()
        self.email_line_edit.setPlaceholderText("example@domain.com")
        self.registerField(WizardField.EMAIL.required(), self.email_line_edit)
        self.email_line_edit.textChanged.connect(self.update_login_button_enabledness)

        self.password_line_edit = QLineEdit()
        self.password_line_edit.setEchoMode(QLineEdit.EchoMode.Password)
        self.password_line_edit.setPlaceholderText("*" * 12)
        self.registerField(WizardField.PASSWORD.required(), self.password_line_edit)
        self.password_line_edit.textChanged.connect(self.update_login_button_enabledness)

        self.peek_button = QToolButton()
        self.peek_button.setIcon(icons.EYE_OPEN())
        self.peek_button.setCheckable(True)
        self.peek_button.setChecked(False)
        self.peek_button.toggled.connect(self._handle_peek_state_change)

        password_and_peek_layout = QHBoxLayout()
        password_and_peek_layout.addWidget(self.password_line_edit)
        password_and_peek_layout.addWidget(self.peek_button)

        remember_me_checkbox = QCheckBox()
        remember_me_checkbox.setText("Remember me")
        remember_me_checkbox.setChecked(False)
        remember_me_terms_label = QLabel(self._REMEMBER_ME_TERMS_HTML)
        remember_me_terms_label.setStyleSheet(f"QLabel {{ color: {icons.TEXT_LIGHTGREY}; }}")
        self.registerField(WizardField.REMEMBER_ME, remember_me_checkbox)

        self.register_btn = QPushButton(
            "Don't have an account? Register on developer.acconeer.com!"
        )
        self.register_btn.setIcon(icons.EXTERNAL_LINK())
        self.register_btn.clicked.connect(lambda: QDesktopServices.openUrl(DEVSITE_REGISTER_URL))

        self.login_button = QPushButton(self._LOGIN_BUTTON_DEFAULT_TEXT)
        self.login_button.setEnabled(False)
        self.login_button.clicked.connect(self.put_login_bg_task)

        layout = QFormLayout()

        layout.addRow(self.feedback_lbl)
        layout.addRow("Email", self.email_line_edit)
        layout.addRow("Password", password_and_peek_layout)
        layout.addRow(remember_me_checkbox, remember_me_terms_label)

        layout.addRow(self.login_button)

        layout.addRow(_Spacer())
        layout.addRow(self.register_btn)

        self.setLayout(layout)

    def update_login_button_enabledness(self) -> None:
        should_be_enabled = (
            self.field(WizardField.EMAIL) != "" and self.field(WizardField.PASSWORD) != ""
        )
        self.login_button.setEnabled(should_be_enabled)

    def put_login_bg_task(self) -> None:
        email = self.field(WizardField.EMAIL)
        pwd = self.field(WizardField.PASSWORD)
        self.login_button.setText(self._LOGIN_BUTTON_PENDING_TEXT)
        self.login_button.setEnabled(False)
        BGTask.create_task(
            login,
            args=(email, pwd),
            return_cb=self.handle_login_return,
            exception_cb=self.handle_login_exception,
        )

    def handle_login_exception(self, exception: Exception) -> None:
        self.login_button.setText(self._LOGIN_BUTTON_DEFAULT_TEXT)
        self.login_button.setEnabled(True)

        error_msg = get_internet_related_error_message(exception)

        self.setField(WizardField.LOGIN_FEEDBACK, error_msg)
        self.setField(WizardField.LOGIN_FEEDBACK_VISIBLE, True)
        self.setField(WizardField.LOGIN_FEEDBACK_STYLESHEET, FeedbackStyleSheet.ERROR)
        self.login_button.setText(self._LOGIN_BUTTON_DEFAULT_TEXT)

    def handle_login_return(self, cookies: requests.cookies.CookieJar) -> None:
        self.login_button.setEnabled(True)

        if cookies is None:
            login_success = False
            session = None
        else:
            (login_success, session, _) = get_content(cookies)

        if login_success:
            obf_email = try_get_obfuscated_email_from_cookiejar(cookiejar=cookies)
            if obf_email is None:
                self.login_button.setText("Logged in")
            else:
                self.login_button.setText(f"Logged in as {obf_email}")

            w = self.wizard()
            assert isinstance(w, FlashWizard)
            w.set_cookies(cookies)
            w.set_session(session)
            self.completeChanged.emit()
            w.next()  # Go to next page automatically
        else:
            self.setField(WizardField.LOGIN_FEEDBACK, "Wrong email or password. Please try again.")
            self.setField(WizardField.LOGIN_FEEDBACK_VISIBLE, True)
            self.setField(WizardField.LOGIN_FEEDBACK_STYLESHEET, FeedbackStyleSheet.ERROR)
            self.login_button.setText(self._LOGIN_BUTTON_DEFAULT_TEXT)

    def _handle_peek_state_change(self, peek_checked: bool) -> None:
        if peek_checked:
            self.password_line_edit.setEchoMode(QLineEdit.EchoMode.Normal)
            self.peek_button.setIcon(icons.EYE_CLOSED())
        else:
            self.password_line_edit.setEchoMode(QLineEdit.EchoMode.Password)
            self.peek_button.setIcon(icons.EYE_OPEN())

    def isComplete(self) -> bool:
        w = self.wizard()
        assert isinstance(w, FlashWizard)
        return w.cookies is not None

    def validatePage(self) -> bool:
        if self.field(WizardField.REMEMBER_ME):
            w = self.wizard()
            assert isinstance(w, FlashWizard)
            save_cookies(w.cookies)

        return True


class LicensePage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.license_has_been_fetched = False

        self.setTitle(DEV_LICENSE_DEFAULT_HEADER)
        self.setButtonText(QWizard.WizardButton.NextButton, "I Accept")

        self.license_text_edit = QTextEdit()
        self.license_text_edit.setReadOnly(True)
        self.license_text_edit.setText("Downloading license ...")
        BGTask.create_task(self.get_dev_licence, args=(), return_cb=self.populate_devlicence)

        layout = QVBoxLayout()
        layout.addWidget(self.license_text_edit)
        self.setLayout(layout)

    def isComplete(self) -> bool:
        return self.license_has_been_fetched

    @staticmethod
    def get_dev_licence() -> tuple[str, str, str]:
        dev_license = DevLicense()
        dev_license.load()

        title = dev_license.get_header()
        subtitle = dev_license.get_subheader_element()
        paragraphs = "".join(dev_license.get_content_elements())

        return (title, subtitle, paragraphs)

    def populate_devlicence(self, t: tuple[str, str, str]) -> None:
        self.license_has_been_fetched = True
        self.completeChanged.emit()

        (title, subtitle, paragraphs) = t
        self.setTitle(title)
        self.license_text_edit.setHtml(subtitle + paragraphs)


class DownloadPage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.setTitle("Download")

        # The flow is weird if going back to download from e.g. flash page.
        # Making this a "commit page" makes it impossible to go back here
        # (and the user needs to cancel the wizard instead)
        self.setCommitPage(True)
        self.setButtonText(QWizard.WizardButton.CommitButton, "Next")

        self._download_button = _ThreeDotAnimatedQPushButton("Download")
        self._download_button.clicked.connect(self.start_download)

        self.feedback_lbl = QLineEdit("")
        self.feedback_lbl.setReadOnly(True)
        self.feedback_lbl.setVisible(False)
        self.feedback_lbl.setStyleSheet(FeedbackStyleSheet.ERROR)

        self._auth_lbl = QLabel("")

        self._forget_me_btn = QPushButton("Forget me")
        self._forget_me_btn.setFlat(True)
        self._forget_me_btn.clicked.connect(self.forget_me)

        layout = QGridLayout()
        layout.addWidget(self._download_button, 0, 0, 1, 2)
        layout.addWidget(self.feedback_lbl, 1, 0, 1, 2)
        layout.addWidget(_Spacer(), 2, 0, 1, 2)
        layout.addWidget(self._auth_lbl, 3, 0, 1, 1)
        layout.addWidget(self._forget_me_btn, 3, 1, 1, 1)

        self.setLayout(layout)

    def initializePage(self) -> None:
        self._download_button.setText(f"Download image for {self.field(WizardField.MODULE)}")
        self._download_button.setEnabled(True)

        w = self.wizard()
        assert isinstance(w, FlashWizard)
        cookies = w.cookies

        if cookies is None:
            self._forget_me_btn.setVisible(False)
        else:
            self._forget_me_btn.setVisible(True)
            mby_obf_email = try_get_obfuscated_email_from_cookiejar(cookies)
            if mby_obf_email is None:
                self._auth_lbl.setText("Logged in.")
            else:
                self._auth_lbl.setText(f"Logged in as {mby_obf_email}")

    def isComplete(self) -> bool:
        return len(self.field(WizardField.BIN_PATH)) > 0

    def forget_me(self) -> None:
        clear_cookies()
        w = self.wizard()
        assert isinstance(w, FlashWizard)
        w.cookies = None
        w.setCurrentId(_Page.LOGIN)

    def start_download(self) -> None:
        self._download_button.setEnabled(False)
        self._download_button.set_text_with_dots_animation("Downloading")

        try:
            w = self.wizard()
            assert isinstance(w, FlashWizard)
            cookies = w.cookies
            session = w.session
            comm_device = w.comm_device

            if cookies is None or comm_device is None:
                # Something has gone wrong, ask user to retry
                w.setCurrentId(_Page.OOPSIE)
                return

            if session is None:
                (login_success, session, _) = get_content(cookies)
                w.session = session
                if not login_success:
                    self.setField(
                        WizardField.LOGIN_FEEDBACK, "Cookies have gone stale, Please login again"
                    )
                    self.setField(WizardField.LOGIN_FEEDBACK_VISIBLE, True)
                    self.setField(WizardField.LOGIN_FEEDBACK_STYLESHEET, FeedbackStyleSheet.NICE)
                    clear_cookies()
                    w.cookies = None
                    w.setCurrentId(_Page.LOGIN)
                    return

            device = get_flash_download_name(
                comm_device,
                device_name=self.field(WizardField.MODULE),
            )
            download_dir_path = get_temp_dir()

            BGTask.create_task(
                download,
                (session, cookies, download_dir_path, device),
                return_cb=self.handle_successful_download,
                exception_cb=self.handle_exception_during_download,
            )
        except Exception:
            self.wizard().setCurrentId(_Page.OOPSIE)

    def handle_successful_download(self, bin_path_and_version: tuple[str, str]) -> None:
        self._download_button.setEnabled(False)

        (bin_path, version) = bin_path_and_version
        self.setField(WizardField.BIN_PATH, bin_path)
        self._download_button.setText(f"Downloaded {version}")
        self.completeChanged.emit()
        self.wizard().next()  # automatically go to next page

    def handle_exception_during_download(self, exception: Exception) -> None:
        self._download_button.setEnabled(True)
        self._download_button.setText("Download")

        self.feedback_lbl.setVisible(True)
        error_msg = get_internet_related_error_message(exception)
        self.feedback_lbl.setText(error_msg)


class FlashPage(QWizardPage):
    def __init__(
        self,
        app_model: AppModel,
    ) -> None:
        super().__init__()

        self.setTitle("Flash")

        self.app_model = app_model
        self.flash_completed_successfully = False

        bin_path_line_edit = QLineEdit()
        bin_path_line_edit.setReadOnly(True)
        self.registerField(WizardField.BIN_PATH.required(), bin_path_line_edit)

        self.feedback_lbl = QLineEdit()
        self.feedback_lbl.setReadOnly(True)
        self.feedback_lbl.setVisible(False)
        self.feedback_lbl.setStyleSheet(f"background-color: {icons.ERROR_RED}; color: white")

        self._flash_button = _ThreeDotAnimatedQPushButton("Flash")
        self._flash_button.clicked.connect(self.flash)

        layout = QFormLayout()
        layout.addRow("Binary path", bin_path_line_edit)
        layout.addRow(self._flash_button)
        layout.addRow(self.feedback_lbl)
        layout.addRow(_Spacer())
        self.setLayout(layout)

    def isComplete(self) -> bool:
        return self.flash_completed_successfully

    def flash(self) -> None:
        if self.app_model.connection_state is not ConnectionState.DISCONNECTED:
            retcode = QMessageBox.warning(
                self,
                "Sensor Connected",
                "Exploration Tool is connected to a sensor that needs to be disconnect before flashing.\n\nDisconnect the sensor?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )

            user_accepted = (retcode & QMessageBox.StandardButton.Yes) > 0
            if user_accepted:
                self.app_model.disconnect_client()
                _ = QMessageBox.information(
                    self,
                    "Sensor Disconnecting ...",
                    "Disconnecting the sensor. Press the <b>Flash</b> button again.",
                    QMessageBox.StandardButton.Ok,
                    QMessageBox.StandardButton.Ok,
                )
            return

        w = self.wizard()
        assert isinstance(w, FlashWizard)
        comm_device = w.comm_device

        if comm_device is None:
            w.setCurrentId(_Page.OOPSIE)
            return

        boot_description = get_boot_description(comm_device, self.field(WizardField.MODULE))
        if boot_description is not None:
            dialog = UserMessageDialog(
                "Bootloader description",
                boot_description,
                "Got it! The board is in bootloader mode",
            )
            diagcode = dialog.exec()
            if diagcode == QDialog.DialogCode.Rejected:
                return

        self._flash_button.set_text_with_dots_animation("Flashing")
        self._flash_button.setEnabled(False)
        self.app_model.port_updater_enable(False)

        BGTask.create_task(
            flash_image,
            args=(self.field(WizardField.BIN_PATH), comm_device, self.field(WizardField.MODULE)),
            return_cb=self.handle_flash_return,
            exception_cb=self.handle_flash_exception,
        )

    def handle_flash_return(self, _: t.Any) -> None:
        self.flash_completed_successfully = True
        self.completeChanged.emit()

        self._flash_button.setText("Flashed!")
        self._flash_button.setEnabled(False)

        self.feedback_lbl.setVisible(False)
        self.app_model.port_updater_enable(True)

        w = self.wizard()
        assert isinstance(w, FlashWizard)
        comm_device = w.comm_device

        if comm_device is None:
            w.setCurrentId(_Page.OOPSIE)
            return

        post_dfu_description = get_post_dfu_description(
            comm_device, self.field(WizardField.MODULE)
        )
        if post_dfu_description is not None:
            dialog = UserMessageDialog(
                "Bootloader description",
                post_dfu_description,
                "Done",
            )
            dialog.exec()

        self.wizard().next()  # automatically go to next page

    def handle_flash_exception(self, exception: Exception) -> None:
        self._flash_button.setText("Flash")
        self._flash_button.setEnabled(True)

        self.feedback_lbl.setVisible(True)
        self.feedback_lbl.setText("Flash failed. Please try again.")

        self.app_model.port_updater_enable(True)


class FinalPage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.setTitle("All done!")

        help_text = "\n".join(
            [
                "If you run into any issues directly after flashing your module,",
                "try removing its power and plugging it back in again.",
            ]
        )

        layout = QVBoxLayout()
        layout.addWidget(QLabel(help_text))
        self.setLayout(layout)


class OopsiePage(QWizardPage):
    def __init__(self) -> None:
        super().__init__()

        self.setTitle("Error")

        github_et_issues_url = "https://github.com/acconeer/acconeer-python-exploration/issues"
        github_et_issues_link = (
            f"<a href='{github_et_issues_url}'>Exploration Tool's GitHub Issues</a>"
        )
        help_text = "<br><br>".join(
            [
                "<b>An unforseen error has occured!</b>",
                "Please restart the flashing process from the beginning.",
                "Use a reliable internet connection if applicable and available.",
                f"If all else fails, please get in touch with us at {github_et_issues_link}",
            ]
        )
        label = QLabel(help_text)
        label.setWordWrap(True)

        layout = QVBoxLayout()
        layout.addWidget(label)
        self.setLayout(layout)


class _Page(IntEnum):
    MODULE_SELECTION = auto()
    DEVICE_SELECTION = auto()
    LOCAL_FILE_OR_DOWNLOAD = auto()
    BROWSE_LOCAL_FILE = auto()
    LICENSE = auto()
    LOGIN = auto()
    DOWNLOAD = auto()
    FLASH = auto()
    FINAL = auto()
    OOPSIE = auto()


class FlashWizard(QWizard):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self.setWindowTitle("Flash Wizard")

        self.comm_device: t.Optional[CommDevice] = None
        self.cookies: t.Optional[requests.cookies.CookieJar] = get_cookies()
        self.session: t.Optional[requests.Session] = None

        self.setPage(_Page.MODULE_SELECTION, ModuleSelectionPage())
        self.setPage(_Page.DEVICE_SELECTION, DeviceSelectionPage(app_model))
        self.setPage(_Page.LOCAL_FILE_OR_DOWNLOAD, SelectLocalFileOrDownloadFromDevsitePage())
        self.setPage(_Page.BROWSE_LOCAL_FILE, BrowseLocalFilePage())
        self.setPage(_Page.LICENSE, LicensePage())
        self.setPage(_Page.LOGIN, LoginFormPage())
        self.setPage(_Page.DOWNLOAD, DownloadPage())
        self.setPage(_Page.FLASH, FlashPage(app_model))
        self.setPage(_Page.FINAL, FinalPage())

        self.setPage(_Page.OOPSIE, OopsiePage())

    def set_comm_device(self, comm_device: CommDevice) -> None:
        self.comm_device = comm_device

    def set_cookies(self, cookies: requests.cookies.CookieJar) -> None:
        self.cookies = cookies

    def set_session(self, session: requests.Session) -> None:
        self.session = session

    def nextId(self) -> int:
        if self.cookies is None:
            license_next_page = _Page.LOGIN
        else:
            license_next_page = _Page.DOWNLOAD

        if self.field(WizardField.DOWNLOAD_SELECTED):
            local_file_or_download_next_page = _Page.LICENSE
        else:
            local_file_or_download_next_page = _Page.BROWSE_LOCAL_FILE

        transitions: dict[int, int] = {
            _Page.MODULE_SELECTION: _Page.DEVICE_SELECTION,
            _Page.DEVICE_SELECTION: _Page.LOCAL_FILE_OR_DOWNLOAD,
            _Page.LOCAL_FILE_OR_DOWNLOAD: local_file_or_download_next_page,
            _Page.LICENSE: license_next_page,
            _Page.LOGIN: _Page.DOWNLOAD,
            _Page.DOWNLOAD: _Page.FLASH,
            _Page.BROWSE_LOCAL_FILE: _Page.FLASH,
            _Page.FLASH: _Page.FINAL,
            _Page.FINAL: -1,
            _Page.OOPSIE: -1,
        }
        return transitions[self.currentId()]


_T = t.TypeVar("_T")


@dataclass
class _AnimationState:
    timer_id: int
    text_stem: str
    frame_no: int

    def increment_frame_no(self) -> None:
        self.frame_no += 1

    def text_to_display(self) -> str:
        dots = "." + ("." * (self.frame_no % 3))
        return self.text_stem + " " + dots.ljust(3)


class _ThreeDotAnimatedQPushButton(QPushButton):
    def __init__(self, text: str) -> None:
        super().__init__(text)
        self._state: t.Optional[_AnimationState] = None

    def setText(self, text: str) -> None:
        self._internal_set_text(text)

        if self._state is None:
            return

        self.killTimer(self._state.timer_id)
        self._state = None

    def _internal_set_text(self, text: str) -> None:
        super().setText(text)

    def set_text_with_dots_animation(self, text: str) -> None:
        self._state = _AnimationState(
            timer_id=self.startTimer(1000),
            text_stem=text,
            frame_no=0,
        )
        self._internal_set_text(self._state.text_to_display())

    def timerEvent(self, ev: QTimerEvent) -> None:
        if self._state is None:
            return

        btn_text = self._state.text_to_display()
        self._state.increment_frame_no()

        self._internal_set_text(btn_text)


def reraise(e: Exception) -> None:
    raise e


class BGTask(QRunnable):
    class Signals(QObject):
        sig_return = Signal(object)  # _T
        sig_exception = Signal(Exception)

    @classmethod
    def create_task(
        cls,
        f: t.Callable[..., _T],
        args: tuple[t.Any, ...],
        return_cb: t.Callable[[_T], t.Any] = lambda _: None,
        exception_cb: t.Callable[[Exception], t.Any] = reraise,
    ) -> None:
        signals = BGTask.Signals()
        signals.sig_return.connect(return_cb)
        signals.sig_exception.connect(exception_cb)

        QThreadPool.globalInstance().start(BGTask(f, args, signals))

    def __init__(self, f: t.Callable[..., _T], args: tuple[t.Any, ...], signals: Signals) -> None:
        super().__init__()
        self.signals = signals
        self.f = f
        self.args = args

    def run(self) -> None:
        try:
            res = self.f(*self.args)
        except Exception as e:
            self.signals.sig_exception.emit(e)
        else:
            self.signals.sig_return.emit(res)


def get_internet_related_error_message(exception: Exception) -> str:
    if isinstance(exception, requests.exceptions.ConnectionError):
        # there is no internet access
        error_msg = "Could not reach the Developer Site. Do you have an internet connection?"
    else:
        error_msg = "An unexpected error occured:\n" + "".join(exception.args)
    return error_msg


def try_get_obfuscated_email_from_cookiejar(
    cookiejar: t.Optional[requests.cookies.CookieJar],
) -> t.Optional[str]:
    # A cookie in the cookiejar k,v pairs look like this (2025-08-19):
    # k = wordpress_logged_in_XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    # v = firstname-lastnamedomain-com%XXXXXXXXXXXXXXXXXXXXXXXXX...
    # (yes, the "@" is stripped)
    if cookiejar is None:
        return None

    try:
        for k, v in cookiejar.items():
            if "logged_in" in k:
                (mangled_email, *_) = str(v).split("%")
                if mangled_email.count("-") < 2:
                    return (
                        mangled_email[0]
                        + "@".center(len(mangled_email) - 2, "*")
                        + mangled_email[-1]
                    )
                else:
                    leftmost_dash_idx = mangled_email.find("-")
                    rightmost_dash_idx = mangled_email.rfind("-")
                    return (
                        mangled_email[:leftmost_dash_idx]
                        + "@".center(rightmost_dash_idx - leftmost_dash_idx, "*")
                        + mangled_email[rightmost_dash_idx + 1 :]
                    )
    except Exception:
        return None

    return None


class _Spacer(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast

import typing_extensions as te

from acconeer.exptool import a121
from acconeer.exptool._core.communication import Client
from acconeer.exptool.a121._core import utils

from ._message import GeneralMessage, Message, StatusMessage
from ._rate_calc import _RateCalculator, _RateStats


ClientT = TypeVar("ClientT", bound=Client[Any, Any, Any, Any, Any])
ResultT = TypeVar("ResultT")


def _format_warning(s: str) -> str:
    return f'<b style="color: #FD5200;">Warning: {s}</b>'


_CALIBRATION_NEEDED_MESSAGE = _format_warning("Calibration needed - restart")
_DATA_SATURATED_MESSAGE = _format_warning("Data saturated - reduce gain")
_FRAME_DELAYED_MESSAGE = _format_warning("Frame delayed")


class _Extractor(te.Protocol):
    @staticmethod
    def create_rate_calculator(client: Any) -> _RateCalculator: ...

    @staticmethod
    def update_rate_calculator(calculator: _RateCalculator, result: Any) -> _RateStats: ...

    @staticmethod
    def get_status_messages(result: Any) -> list[str]: ...


class _A121Extractor(_Extractor):
    @staticmethod
    def create_rate_calculator(client: a121.Client) -> _RateCalculator:
        assert client.extended_metadata is not None
        assert client.server_info is not None

        (first_metadata, *_) = utils.iterate_extended_structure_values(client.extended_metadata)
        return _RateCalculator(client.server_info.ticks_per_second, first_metadata.tick_period)

    @staticmethod
    def update_rate_calculator(
        calculator: _RateCalculator, result: Union[a121.Result, list[dict[int, a121.Result]]]
    ) -> _RateStats:
        if isinstance(result, a121.Result):
            return calculator.update(result.tick, result.frame_delayed)
        else:
            (first_result, *_) = utils.iterate_extended_structure_values(result)
            return calculator.update(first_result.tick, first_result.frame_delayed)

    @staticmethod
    def get_status_messages(
        result: Union[a121.Result, list[dict[int, a121.Result]]],
    ) -> list[str]:
        if isinstance(result, a121.Result):
            result_list = [result]
        elif isinstance(result, list):
            result_list = list(utils.iterate_extended_structure_values(result))
        else:
            msg = f"Unexpected type for 'result': {type(result)}"
            raise AssertionError(msg)

        msgs = []
        if any(r.data_saturated for r in result_list):
            msgs.append(_DATA_SATURATED_MESSAGE)

        if any(r.calibration_needed for r in result_list):
            msgs.append(_CALIBRATION_NEEDED_MESSAGE)

        if any(r.frame_delayed for r in result_list):
            msgs.append(_FRAME_DELAYED_MESSAGE)

        return msgs


class ApplicationClient(Generic[ClientT]):
    callback: Callable[[Message], None]
    extractor_strategy: type[_Extractor]
    _wrapped_client: Client[Any, Any, Any, Any, Any]
    _rate_stats_calc: Optional[_RateCalculator]
    _frame_count: int

    def __init__(
        self,
        wrapped_client: ClientT,
        callback: Callable[[Message], None],
        extractor_strategy: type[_Extractor],
    ) -> None:
        self._wrapped_client = wrapped_client
        self.callback = callback
        self.extractor_strategy = extractor_strategy
        self._rate_stats_calc = None
        self._frame_count = 0

    def __getattr__(self, name: str) -> Any:
        """
        This is the key to the "fall-through-wrapping".
        """
        return getattr(self._wrapped_client, name)

    @classmethod
    def wrap_a121(cls, wrapped_client: ClientT, callback: Callable[[Message], None]) -> ClientT:
        """
        Factory that makes the ApplicationClient transparent from a
        typing perspective.

        The cast is necessary as we are dealing with magic here, but should hold
        up so long any overriden functions have the same type as the overriders.
        """
        return cast(ClientT, cls(wrapped_client, callback, _A121Extractor))

    def start_session(self) -> None:
        self._wrapped_client.start_session()
        self._rate_stats_calc = self.extractor_strategy.create_rate_calculator(
            self._wrapped_client
        )

    def get_next(self) -> Any:
        assert self._rate_stats_calc is not None

        result = self._wrapped_client.get_next()
        stats = self.extractor_strategy.update_rate_calculator(self._rate_stats_calc, result)

        self.callback(GeneralMessage(name="rate_stats", data=stats))
        self._frame_count += 1
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))

        for msg in self.extractor_strategy.get_status_messages(result):
            self.callback(StatusMessage(status=msg))

        return result

    def stop_session(self) -> None:
        result = self._wrapped_client.stop_session()
        self._frame_count = 0
        self.callback(GeneralMessage(name="rate_stats", data=None))
        self.callback(GeneralMessage(name="frame_count", data=None))
        return result



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import traceback
from typing import Callable, Dict, Optional

from ._message import LogMessage, Message


class BackendLoggerError(Exception):
    pass


class BackendLogger:
    _callback: Optional[Callable[[Message], None]] = None
    _instances: Dict[str, BackendLogger] = {}

    def __init__(self, name: str):
        self.name = name

    def _log(self, log_level: str, log_string: str) -> None:
        if not BackendLogger._callback:
            msg = "BackendLogger not initialized"
            raise BackendLoggerError(msg)
        BackendLogger._callback(
            LogMessage(module_name=self.name, log_level=log_level, log_string=log_string)
        )

    def critical(self, log_string: str) -> None:
        self._log("CRITICAL", log_string)

    def error(self, log_string: str) -> None:
        self._log("ERROR", log_string)

    def exception(self, exception: BaseException) -> None:
        tb = "\n".join(traceback.format_tb(exception.__traceback__))
        self._log(
            "ERROR",
            f"An exception was raised:\n{tb}{type(exception).__name__}: {exception}",
        )

    def warning(self, log_string: str) -> None:
        self._log("WARNING", log_string)

    def info(self, log_string: str) -> None:
        self._log("INFO", log_string)

    def debug(self, log_string: str) -> None:
        self._log("DEBUG", log_string)

    @staticmethod
    def set_callback(callback: Callable[[Message], None]) -> None:
        BackendLogger._callback = callback

    @staticmethod
    def getLogger(name: str) -> BackendLogger:
        logger = BackendLogger._instances.get(name)
        if not logger:
            logger = BackendLogger(name)
            BackendLogger._instances[name] = logger
            logger.debug(f"BackendLogger for {name} initialized")
        return logger



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel, QSplitter, QVBoxLayout, QWidget

from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui.icons import BUTTON_ICON_COLOR
from acconeer.exptool.app.new.ui.misc import VerticalSeparator
from acconeer.exptool.app.new.ui.utils import LayoutWrapper

from .connection_widget import ClientConnectionWidget, GenerationSelection
from .hints import HintWidget
from .plugin_widget import PluginControlArea, PluginPlotArea, PluginSelectionArea
from .recording_widget import RecordingWidget


class StreamingMainWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        top_bar = TopBar(app_model, self)
        layout.addWidget(top_bar)
        layout.setStretchFactor(top_bar, 0)

        working_area = WorkingArea(app_model, self)
        layout.addWidget(working_area)
        layout.setStretchFactor(working_area, 1)

        self.setLayout(layout)


class TopBar(QFrame):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setObjectName("TopBar")
        self.setStyleSheet("QFrame#TopBar {background: #ebebeb;}")

        layout = QHBoxLayout(self)

        layout.addWidget(GenerationSelection(app_model, self))
        layout.addWidget(VerticalSeparator(self))
        layout.addWidget(ClientConnectionWidget(app_model, self))
        layout.addWidget(HintWidget(app_model, self))
        layout.addStretch(1)
        layout.addWidget(RecordingWidget(app_model, self))

        self.setLayout(layout)


class PluginTitleLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__("", parent)

        self.setStyleSheet(f"color: {BUTTON_ICON_COLOR}; text-align: center; font-weight: bold;")
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        app_model.sig_notify.connect(self.on_app_model_update)

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.setText(app_model.plugin.title if app_model.plugin is not None else "")


class WorkingArea(QSplitter):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        layout = QHBoxLayout(self)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        left_area = PluginSelectionArea(app_model, self)
        left_area.setMinimumWidth(250)
        left_area.setMaximumWidth(350)
        layout.addWidget(left_area)
        layout.setStretchFactor(left_area, 0)

        plugin_layout = QVBoxLayout()

        plugin_label = PluginTitleLabel(app_model, self)
        plugin_layout.addWidget(plugin_label)
        plugin_layout.setStretchFactor(plugin_label, 0)

        plot_plugin_area = PluginPlotArea(app_model, self)
        plugin_layout.addWidget(plot_plugin_area)
        plugin_layout.setStretchFactor(plot_plugin_area, 1)

        plot_layout_wrapper = LayoutWrapper(plugin_layout)
        layout.addWidget(plot_layout_wrapper)
        layout.setStretchFactor(plot_layout_wrapper, 1)

        right_area = PluginControlArea(app_model, self)
        right_area.setMinimumWidth(400)
        right_area.setMaximumWidth(450)
        layout.addWidget(right_area)
        layout.setStretchFactor(right_area, 0)

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import numpy as np
import typing_extensions as te

from acconeer.exptool._core.communication import Message, ParseError
from acconeer.exptool.a121._core.entities import Metadata, SensorCalibration


class SensorCalibrationDict(te.TypedDict):
    temperature: int
    data: str


class MetadataDict(te.TypedDict):
    frame_data_length: int
    sweep_data_length: int
    subsweep_data_offset: list[int]
    subsweep_data_length: list[int]
    calibration_temperature: int
    base_step_length_m: float
    max_sweep_rate: float


class SetupResponseHeader(te.TypedDict):
    tick_period: int
    metadata: list[list[MetadataDict]]
    sensor_calibrations: t.Optional[dict[int, SensorCalibrationDict]]


@attrs.frozen
class SetupResponse(Message):
    grouped_metadatas: list[list[Metadata]]
    sensor_calibrations: t.Optional[dict[int, SensorCalibration]]

    @classmethod
    def parse(cls, header: dict[str, t.Any], payload: bytes) -> SetupResponse:
        t.cast(SetupResponseHeader, header)

        try:
            metadata_groups = header["metadata"]
            calibration_info_list = header.get("calibration_info")

            metadata = [
                [
                    Metadata(
                        frame_data_length=metadata_dict["frame_data_length"],
                        sweep_data_length=metadata_dict["sweep_data_length"],
                        subsweep_data_offset=np.array(metadata_dict["subsweep_data_offset"]),
                        subsweep_data_length=np.array(metadata_dict["subsweep_data_length"]),
                        calibration_temperature=metadata_dict["calibration_temperature"],
                        base_step_length_m=metadata_dict["base_step_length_m"],
                        max_sweep_rate=metadata_dict["max_sweep_rate"],
                        high_speed_mode=metadata_dict.get("high_speed_mode", None),
                        tick_period=header["tick_period"],
                    )
                    for metadata_dict in metadata_group
                ]
                for metadata_group in metadata_groups
            ]

            sensor_calibrations = None
            if calibration_info_list:
                sensor_calibrations = {
                    calibration_info_dict["sensor_id"]: SensorCalibration(
                        temperature=calibration_info_dict["temperature"],
                        data=calibration_info_dict["data"],
                    )
                    for calibration_info_dict in calibration_info_list
                }
            return cls(metadata, sensor_calibrations)

        except KeyError as ke:
            raise ParseError from ke



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import abc
import functools
from typing import Any, Callable, Iterator, TypeVar

import numpy as np
import numpy.typing as npt
from typing_extensions import Concatenate, ParamSpec

from acconeer.exptool._core.entities.client_info import ClientInfo
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities.configs import SessionConfig

from .metadata import Metadata
from .result import Result
from .sensor_calibration import SensorCalibration
from .server_info import ServerInfo
from .stacked_results import StackedResults


_T = TypeVar("_T")
_P = ParamSpec("_P")


class RecordException(Exception):
    pass


class SessionRecord(abc.ABC):
    """Record representing all data needed to recreate a sinle session"""

    @property
    @abc.abstractmethod
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """Extended metadata"""

    @property
    def metadata(self) -> Metadata:
        """Retrieves the sole metadata entry in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_metadata`
        """

        return utils.unextend(self.extended_metadata)

    @property
    @abc.abstractmethod
    def extended_results(self) -> Iterator[list[dict[int, Result]]]:
        """The sequence of extended results

        Retrieves an iterator over all extended results in the record, which may be useful to
        replay a session.

        .. tip::

            If your record only has one entry in the session, and you want to retrieve the results
            for that one entry, you can use :attr:`results`.

            If you desire to get, for example, all frames in the sequence of results stacked in one
            array, use
            :attr:`extended_stacked_results`. Similar to :attr:`results`, there is also
            :attr:`stacked_results` which retrieves the one and only entry.
        """

    @property
    def results(self) -> Iterator[Result]:
        """Retrieves the sole sequence of results in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_results`
        """

        for extended_result in self.extended_results:
            yield utils.unextend(extended_result)

    @property
    @abc.abstractmethod
    def extended_stacked_results(self) -> list[dict[int, StackedResults]]:
        """The extended stacked results"""

    @property
    def stacked_results(self) -> StackedResults:
        """Retrieves the sole stacked results in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_stacked_results`
        """

        return utils.unextend(self.extended_stacked_results)

    @property
    def frames(self) -> npt.NDArray[np.complex128]:
        """Retrieves the sole stack of frames in the record

        Alias for ``Record.stacked_results.frame``.

        :raises: ValueError if there are multiple entries in the :attr:`extended_stacked_results`
        """

        return self.stacked_results.frame

    @property
    @abc.abstractmethod
    def num_frames(self) -> int:
        """The number of frames in the record"""

    @property
    @abc.abstractmethod
    def session_config(self) -> SessionConfig:
        """Session config"""

    @property
    @abc.abstractmethod
    def sensor_id(self) -> int:
        """The sole sensor ID

        :raises: ValueError if there are multiple entries in the :attr:`session_config`
        """

    @property
    @abc.abstractmethod
    def calibrations(self) -> dict[int, SensorCalibration]:
        """The calibrations in the record"""

    @property
    @abc.abstractmethod
    def calibrations_provided(self) -> dict[int, bool]:
        """The calibrations provided in the record"""


def sole_accessor(
    f: Callable[Concatenate[Record, _P], _T],
) -> Callable[Concatenate[Record, _P], _T]:
    @functools.wraps(f)
    def decorated(instance: Record, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        num_sessions = instance.num_sessions
        name = f.__name__
        if num_sessions > 1:
            raise ValueError(
                f"Cannot access sole '{name}' since there are {num_sessions} sessions. "
                + f"Try accessing a specific session's '{name}' with '.session(<index>).{name}'"
            )
        return f(instance, *args, **kwargs)

    return decorated


class Record(SessionRecord):
    """
    Record representing all data needed to recreate multiple sessions

    A Record also is a SessionRecord, but its session-related accessors
    (like results, session_config) will raise an error if this Record contains
    multiple sessions/:class:`SessionRecords`.
    """

    @property
    @abc.abstractmethod
    def client_info(self) -> ClientInfo:
        """Client info"""

    @property
    @abc.abstractmethod
    def lib_version(self) -> str:
        """The version of the ``acconeer.exptool`` library which created the record"""

    @property
    @abc.abstractmethod
    def server_info(self) -> ServerInfo:
        """Server info"""

    @property
    @abc.abstractmethod
    def timestamp(self) -> str:
        """Creation timestamp"""

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        """UUID"""

    @abc.abstractmethod
    def session(self, session_index: int) -> SessionRecord:
        """Gets the session at session_index"""

    @property
    @abc.abstractmethod
    def num_sessions(self) -> int:
        "The number of sessions in this record"

    @property
    @sole_accessor
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """The extended metadata of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).extended_metadata

    @property
    @sole_accessor
    def extended_results(self) -> Iterator[list[dict[int, Result]]]:
        """The extended results of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).extended_results

    @property
    @sole_accessor
    def extended_stacked_results(self) -> list[dict[int, StackedResults]]:
        """The extended stacked results of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).extended_stacked_results

    @property
    @sole_accessor
    def num_frames(self) -> int:
        """The number of frames of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).num_frames

    @property
    @sole_accessor
    def session_config(self) -> SessionConfig:
        """The session config of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).session_config

    @property
    @sole_accessor
    def sensor_id(self) -> int:
        """The sole session id of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).sensor_id

    @property
    @sole_accessor
    def calibrations(self) -> dict[int, SensorCalibration]:
        """The calibrations of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).calibrations

    @property
    @sole_accessor
    def calibrations_provided(self) -> dict[int, bool]:
        """A mapping of reused (provided) calibrations of the sole session in this Record

        :raises: ValueError if this record contains multiple sessions
        """
        return self.session(0).calibrations_provided


class PersistentRecord(Record):
    """Record that wraps a file on disk

    Data is lazily loaded from the underlying file on demand.
    """

    @abc.abstractmethod
    def close(self) -> None:
        """Close the underlying file"""

    def __enter__(self) -> Record:
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo._utils import estimate_frame_rate
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetGroupFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget
from acconeer.exptool.app.new.ui.components.a121 import (
    RangeHelpView,
    SensorConfigEditor,
)
from acconeer.exptool.app.new.ui.components.json_save_load_buttons import (
    JsonButtonOperations,
)
from acconeer.exptool.app.new.ui.stream_tab.plugin_widget import PluginPlotArea

from ._configs import get_default_config, get_traffic_config
from ._detector import (
    Detector,
    DetectorConfig,
    DetectorMetadata,
    DetectorResult,
    _load_algo_data,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: DetectorConfig = attrs.field(factory=DetectorConfig)


class PluginPresetId(Enum):
    DEFAULT = auto()
    TRAFFIC = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    detector_config: DetectorConfig
    detector_metadata: DetectorMetadata
    estimated_frame_rate: float
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: get_default_config(),
        PluginPresetId.TRAFFIC.value: lambda: get_traffic_config(),
    }

    def __init__(
        self,
        callback: Callable[[Message], None],
        generation: PluginGeneration,
        key: str,
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            detector_config=self.shared_state.config,
        )
        sensor_config = self._detector_instance._get_sensor_config(self._detector_instance.config)
        session_config = a121.SessionConfig(
            {self.shared_state.sensor_id: sensor_config},
            extended=False,
        )

        estimated_frame_rate = estimate_frame_rate(self.client, session_config)

        self._detector_instance.start(recorder)
        assert self._detector_instance.detector_metadata is not None
        self.callback(
            SetupMessage(
                detector_config=self.shared_state.config,
                detector_metadata=self._detector_instance.detector_metadata,
                estimated_frame_rate=estimated_frame_rate,
            )
        )

    def end_session(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        self._detector_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._detector_instance is None:
            raise RuntimeError
        result = self._detector_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._is_setup = False
        self._plot_job: Optional[DetectorResult] = None

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.detector_config,
                message.detector_metadata,
                message.estimated_frame_rate,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        detector_config: DetectorConfig,
        detector_metadata: DetectorMetadata,
        estimated_frame_rate: float,
    ) -> None:
        self.plot_layout.clear()

        self.detector_config = detector_config

        self.n_depths = detector_metadata.num_points

        max_update_rate = PluginPlotArea._FPS

        if estimated_frame_rate > max_update_rate:
            plugin_frame_rate = float(max_update_rate)
        else:
            plugin_frame_rate = estimated_frame_rate

        self.history_length_s = 10.0
        self.history_length = int(self.history_length_s * plugin_frame_rate)

        self.time_window_length_s = 3.0
        self.time_window_length_n = int(self.time_window_length_s * plugin_frame_rate)

        self.speed_history = np.zeros(self.history_length)
        self.speed_history_xs = np.array([i for i in range(-self.history_length, 0)])

        n_ticks_to_display = 10
        x_labels = np.linspace(-self.history_length_s, 0, self.history_length)
        all_ticks = [
            (t, "{:.0f}".format(label)) for t, label in zip(self.speed_history_xs, x_labels)
        ]
        subsample_step = self.history_length // n_ticks_to_display
        display_ticks = [all_ticks[::subsample_step]]

        win = self.plot_layout

        # FFT plot

        self.raw_fft_plot = win.addPlot(row=1, col=0)
        self.raw_fft_plot.setTitle("Frequency data")
        self.raw_fft_plot.setLabel(axis="left", text="Amplitude")
        self.raw_fft_plot.setLabel(axis="bottom", text="Frequency", units="Hz")
        self.raw_fft_plot.addLegend(labelTextSize="10pt")
        self.raw_fft_limits = et.utils.SmoothLimits()
        self.raw_fft_plot.setMenuEnabled(False)
        self.raw_fft_plot.setMouseEnabled(x=False, y=False)
        self.raw_fft_plot.hideButtons()
        self.raw_fft_plot.showGrid(x=True, y=True)
        self.raw_fft_plot.setLogMode(x=False, y=True)
        self.raw_fft_curves = []
        self.raw_fft_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.raw_thresholds_curves = []

        for i in range(self.n_depths):
            raw_fft_curve = self.raw_fft_plot.plot(pen=et.utils.pg_pen_cycler(i), name="Fft")
            threshold_curve = self.raw_fft_plot.plot(
                pen=et.utils.pg_pen_cycler(i, style="--"), name="Threshold"
            )
            self.raw_fft_curves.append(raw_fft_curve)
            self.raw_thresholds_curves.append(threshold_curve)

        self.speed_history_plot = win.addPlot(row=0, col=0)
        self.speed_history_plot.setTitle("Speed history")
        self.speed_history_plot.setLabel(axis="left", text="Speed", units="m/s")
        self.speed_history_plot.setLabel(axis="bottom", text="Time", units="Seconds")
        self.speed_history_plot.addLegend(labelTextSize="10pt")
        self.speed_history_curve = self.speed_history_plot.plot(
            pen=None,
            name="speed",
            symbol="o",
            symbolsize=3,
        )

        if detector_config.sweep_rate is not None:
            actual_max_speed = detector_config._get_max_speed(detector_config.sweep_rate)
            self.speed_history_plot.setYRange(-actual_max_speed, actual_max_speed)
        else:
            self.speed_history_plot.setYRange(
                -detector_config.max_speed, detector_config.max_speed
            )
        self.speed_history_plot.setXRange(-self.history_length, 0)
        ay = self.speed_history_plot.getAxis("bottom")
        ay.setTicks(display_ticks)

        self.speed_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )

        self.speed_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0.5),
        )

        self.speed_text_item.setPos(-self.history_length / 2, -detector_config.max_speed / 2)
        brush = et.utils.pg_brush_cycler(1)
        self.speed_history_peak_plot_item = pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=8, symbolBrush=brush, symbolPen="k"
        )
        self.speed_history_plot.addItem(self.speed_history_peak_plot_item)
        self.speed_history_plot.addItem(self.speed_text_item)

        self.speed_text_item.hide()

    def draw_plot_job(self, data: DetectorResult) -> None:
        psd = data.extra_result.psd
        speed_guess = data.max_speed
        x_speeds = data.extra_result.velocities
        thresholds = data.extra_result.actual_thresholds

        self.speed_history = np.roll(self.speed_history, -1)

        self.speed_history[-1] = speed_guess

        if self.time_window_length_n > 0:
            pos_speed = np.max(self.speed_history[-self.time_window_length_n :])
            pos_ind = int(np.argmax(self.speed_history[-self.time_window_length_n :]))
            neg_speed = np.min(self.speed_history[-self.time_window_length_n :])
            neg_ind = int(np.argmin(self.speed_history[-self.time_window_length_n :]))

            if abs(neg_speed) > abs(pos_speed):
                max_display_speed = neg_speed
                max_display_ind = neg_ind
            else:
                max_display_speed = pos_speed
                max_display_ind = pos_ind
        else:
            max_display_speed = self.speed_history[-1]
            max_display_ind = -1

        if max_display_speed != 0.0:
            speed_text = "Max speed estimate {:.4f} m/s".format(max_display_speed)
            speed_html = self.speed_html_format.format(speed_text)

            self.speed_text_item.setHtml(speed_html)
            self.speed_text_item.show()

            sub_xs = self.speed_history_xs[-self.time_window_length_n :]
            self.speed_history_peak_plot_item.setData(
                [sub_xs[max_display_ind]], [max_display_speed]
            )
        else:
            self.speed_history_peak_plot_item.clear()
            self.speed_text_item.hide()

        display_inds = np.array([i for i, x in enumerate(self.speed_history) if x != 0.0])
        if len(display_inds) > 0:
            display_xs = self.speed_history_xs[display_inds]
            display_data = self.speed_history[display_inds]
        else:
            display_xs = []
            display_data = []
        self.speed_history_curve.setData(display_xs, display_data)

        assert psd is not None
        assert thresholds is not None

        top_max = max(np.max(psd), np.max(thresholds))

        smooth_max_val = np.log10(self.raw_fft_smooth_max.update(top_max))
        self.raw_fft_plot.setYRange(-2, smooth_max_val)
        for i in range(psd.shape[1]):
            self.raw_fft_curves[i].setData(x_speeds, psd[:, i])

            threshold_line = np.full(x_speeds.shape[0], thresholds[i])
            self.raw_thresholds_curves[i].setData(x_speeds, threshold_line)


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor[DetectorConfig](
            title="Detector parameters",
            factory_mapping=self._get_pidget_mapping(),
            config_type=DetectorConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)

        self.range_help_view = RangeHelpView()
        scrolly_layout.addWidget(self.range_help_view)

        scrolly_layout.addWidget(self.config_editor)

        self.sensor_config_status = SensorConfigEditor(
            json_button_operations=JsonButtonOperations.SAVE
        )
        self.sensor_config_status.set_read_only(True)
        collapsible_widget = CollapsibleWidget(
            "Current sensor settings", self.sensor_config_status, self.scrolly_widget
        )
        scrolly_layout.addWidget(collapsible_widget)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        service_parameters = {
            "start_point": pidgets.IntPidgetFactory(
                name_label_text="Start point:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "start_point"),
            ),
            "num_points": pidgets.IntPidgetFactory(
                name_label_text="Number of points:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "num_points"),
                limits=(1, 100),
            ),
            "step_length": pidgets.OptionalIntPidgetFactory(
                name_label_text="Step length:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "step_length"),
                checkbox_label_text="Override",
                limits=(1, None),
                init_set_value=72,
            ),
            "profile": pidgets.OptionalEnumPidgetFactory(
                name_label_text="Profile:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "profile"),
                checkbox_label_text="Override",
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (shortest)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (longest)",
                },
            ),
            "sweep_rate": pidgets.OptionalIntPidgetFactory(
                name_label_text="Sweep rate:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "sweep_rate"),
                suffix=" Hz",
                checkbox_label_text="Override",
                limits=(1, 134000),
                init_set_value=10000,
            ),
            "frame_rate": pidgets.OptionalFloatPidgetFactory(
                name_label_text="Frame rate:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "frame_rate"),
                suffix=" Hz",
                checkbox_label_text="Override",
                limits=(1, 200),
                init_set_value=20.0,
            ),
            "hwaas": pidgets.OptionalIntPidgetFactory(
                name_label_text="HWAAS:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "hwaas"),
                checkbox_label_text="Override",
                limits=(1, 511),
                init_set_value=4,
            ),
            "num_bins": pidgets.IntPidgetFactory(
                name_label_text="Number of bins:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "num_bins"),
                limits=(3, 4095),
            ),
        }
        processing_parameters = {
            "max_speed": pidgets.FloatPidgetFactory(
                name_label_text="Max speed:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "max_speed"),
                suffix=" m/s",
                limits=(0, 150.0),
            ),
            "threshold": pidgets.FloatPidgetFactory(
                name_label_text="Detection threshold:",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "threshold"),
                limits=(1.0, 10000.0),
            ),
        }
        return {
            pidgets.FlatPidgetGroup(): service_parameters,
            pidgets.FlatPidgetGroup(): processing_parameters,
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
        else:
            self.config_editor.set_data(state.config)
            self.sensor_id_pidget.set_data(state.sensor_id)

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            sensor_config = Detector._get_sensor_config(state.config)
            self.range_help_view.set_data(sensor_config.subsweep)
            self.sensor_config_status.set_data(sensor_config)

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.config_editor, self.sensor_id_pidget],
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


SPEED_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="speed_detector",
    title="Speed detector",
    docs_link="https://docs.acconeer.com/en/latest/detectors/a121/speed_detector.html",
    description="Measure speed.",
    family=PluginFamily.DETECTOR,
    presets=[
        PluginPresetBase(
            name="Default",
            description="Default settings to validate functionality",
            preset_id=PluginPresetId.DEFAULT,
        ),
        PluginPresetBase(
            name="Traffic",
            description="Settings to validate fast traffic",
            preset_id=PluginPresetId.TRAFFIC,
        ),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from . import visual_policies
from .plot_plugin_base import PgPlotPlugin, PlotPluginBase
from .plugin_preset_base import PluginPresetBase
from .plugin_spec_base import PluginSpecBase
from .view_plugin_base import ViewPluginBase



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import copy
import itertools
from functools import partial
from typing import (
    Any,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import attrs

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool._core.entities.validation_result import Criticality, ValidationResult
from acconeer.exptool.app.new.ui.components.overlay import Overlay
from acconeer.exptool.app.new.ui.utils import LayoutWrapper

from .data_editor import DataEditor
from .group_box import GroupBox
from .json_save_load_buttons import (
    JsonPresentable,
    PresenterFunc,
    create_json_save_load_buttons_from_type,
)
from .pidgets import FlatPidgetGroup, Pidget, PidgetGroup, PidgetHook, WidgetHook
from .types import PidgetFactoryMapping, PidgetGroupFactoryMapping


T = TypeVar("T", bound=JsonPresentable)


def _to_group_factory_mapping(
    factory_mapping: Union[PidgetFactoryMapping, PidgetGroupFactoryMapping],
) -> PidgetGroupFactoryMapping:
    if factory_mapping == {}:
        return {}

    (first_key, *_) = factory_mapping

    # The casts boils down to "non-transferable" type narrowing
    # of a Mapping (typically a dict) given its keys.
    #   In this function, we want to do type narrowing of the parameter type
    # (rewritten as dict[Union[str, PidgetGroup], ...]). The key will have type
    # Union[str, PidgetGroup], which can be narrowed to str or PidgetGroup.
    # This type narrowing is not transferred to the original dict, requiring a cast.
    if isinstance(first_key, PidgetGroup):
        return cast(PidgetGroupFactoryMapping, factory_mapping)

    if isinstance(first_key, str):
        return {FlatPidgetGroup(): cast(PidgetFactoryMapping, factory_mapping)}

    msg = "factory_mapping was neither a PidgetFactoryMappingi nor a PidgetGroupFactoryMapping"
    raise RuntimeError(msg)


class AttrsConfigEditor(DataEditor[Optional[T]]):
    _config: Optional[T]
    _erroneous_aspects: set[str]

    sig_update = Signal(object)

    def __init__(
        self,
        title: str,
        factory_mapping: Union[PidgetFactoryMapping, PidgetGroupFactoryMapping],
        config_type: Type[T],
        *,
        save_load_buttons: bool = True,
        min_top_padding: int = 0,
        extra_presenter: PresenterFunc = lambda i, t: None,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)
        self._config = None
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(11)

        group_box = GroupBox.vertical(
            title,
            right_header=(
                create_json_save_load_buttons_from_type(
                    self, config_type, extra_presenter=extra_presenter
                )
                if save_load_buttons
                else None
            ),
            min_top_padding=min_top_padding,
            parent=self,
        )
        layout.addWidget(group_box)

        self._pidget_mapping: dict[str, Pidget] = {}
        self._pidget_hooks: dict[str, Sequence[PidgetHook]] = {}
        self._group_widgets: list[QWidget] = []
        self._group_hooks: list[Sequence[WidgetHook]] = []
        self._erroneous_aspects = set()

        for pidget_group, factory_mapping in _to_group_factory_mapping(factory_mapping).items():
            pidgets = []
            for aspect, factory in factory_mapping.items():
                pidget = factory.create(group_box)
                pidget.sig_update.connect(partial(self._update_config_aspect, aspect))

                self._pidget_mapping[aspect] = pidget
                self._pidget_hooks[aspect] = factory.hooks

                pidgets.append(pidget)

            group_widget = pidget_group.get_container(pidgets)
            group_box.layout().addWidget(group_widget)

            self._group_widgets.append(group_widget)
            self._group_hooks.append(pidget_group.hooks)

        self.setLayout(layout)

    def set_data(self, config: Optional[T]) -> None:
        if self._config == config:
            return

        self._config = config
        self.setEnabled(config is not None)

        if self._config is None:
            return

        for aspect, pidget in self._pidget_mapping.items():
            if aspect in self._erroneous_aspects:
                continue
            config_value = getattr(self._config, aspect)
            pidget.set_data(config_value)

        for aspect, hooks in self._pidget_hooks.items():
            for hook in hooks:
                hook(self._pidget_mapping[aspect], self._pidget_mapping)

        for group_widget, hooks in zip(self._group_widgets, self._group_hooks):
            for hook in hooks:
                hook(group_widget, self._pidget_mapping)

    def get_data(self) -> Optional[T]:
        return copy.deepcopy(self._config)

    def handle_validation_results(self, results: list[ValidationResult]) -> list[ValidationResult]:
        for aspect, pidget in self._pidget_mapping.items():
            if aspect not in self._erroneous_aspects:
                pidget.set_note_text("")

        unhandled_results: list[ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return self._erroneous_aspects == set()

    def _handle_validation_result(self, result: ValidationResult) -> bool:
        if result.aspect is None:
            return False

        if result.aspect in self._erroneous_aspects:
            # If there is an erroneous aspect in the GUI, we do not want to overwrite that.
            # Once the erroneous aspect is handled with, the same validation result will
            # come through here anyway
            return True

        result_handled = False

        if result.source == self._config:
            for aspect, pidget in self._pidget_mapping.items():
                if result.aspect == aspect:
                    self._pidget_mapping[result.aspect].set_note_text(
                        result.message, result.criticality
                    )
                    result_handled = True

        return result_handled

    def _update_config_aspect(self, aspect: str, value: Any) -> None:
        if self._config is None:
            return

        try:
            # The passing of "self._config: T" is not type safe. We cannot know
            # if "T" is an attrs class in the scope of this class without binding
            # "T" to a common superclass.
            # "AlgoConfigBase" is a candidate, but that is part of the algo-package.
            config = attrs.evolve(self._config, **{aspect: value})  # type: ignore[misc]
        except Exception as e:
            self._erroneous_aspects.add(aspect)
            self._pidget_mapping[aspect].set_note_text(e.args[0], Criticality.ERROR)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._pidget_mapping[aspect].set_note_text(None)
            self._erroneous_aspects.discard(aspect)

            self.set_data(config)
            self.sig_update.emit(config)

    def set_pidget_enabled(self, aspect: str, enabled: bool) -> None:
        if (pidget := self._pidget_mapping.get(aspect)) is None:
            return
        pidget.setEnabled(enabled)


class JoinedPartialAttrsConfigEditors(DataEditor[Optional[T]]):
    """
    A Partial Editor is a pattern that has emerged where
    an ``AttrsConfigEditor`` is constructed with an ``PidgetFactoryMapping``
    that doesn't fully cover all the fields of a config.

    For example: say you want an editor for the following object

        @attrs.mutable
        class Person:
            age: int
            height: int

    Then you might have an editor that looks something like this:

        +- Person physical properties --|S| |L|-+
        |                                       |
        |  Age:                     |_____25_|  |
        |  Height:                  |____170_|  |
        +---------------------------------------+

    with the corresponding pidget mapping

        {"age": IntPidgetFactory(...), "height": IntPidgetFactory(...)}


    This is what the ``AttrsConfigEditor`` class is used for and it also
    supports the save-/load buttons (denoted |S| & |L|).

    However, if now a field like ``favorite_number: int`` was added to the ``Person`` class.
    A field like that doesn't fit the category "physical properties" so you would like to
    separate them visually and have the values stored in the same logical place (the ``Person`` class):

        +- Person physical properties --|S| |L|-+
        |                                       |
        |  Age:                     |_____25_|  |
        |  Height:                  |____170_|  |
        +---------------------------------------+
        +- Person fun facts --------------------+
        |                                       |
        |  Favorite number:         |______7_|  |
        +---------------------------------------+

    The sketch shows 2 partial editors of ``Person``s and that is what this class is used for.
    Programmatically, creating one looks something like this:

        JoinedPartialAttrsConfigEditors(
            {"Person physical properties": {"age": IntPidgetFactory(...), "height": IntPidgetFactory(...)},
             "Person fun facts": {"favorite_number": IntPidgetFactory(...)}},
        )

    The implementation makes sure that set_data/get_data propagates/joins the information
    from each partial widget correctly and also that the save-/load buttons behave as expected.
    """

    def __init__(
        self,
        editor_kwargs: dict[str, Union[PidgetFactoryMapping, PidgetGroupFactoryMapping]],
        config_type: type[T],
        save_load_buttons: bool = True,
        extra_presenter: PresenterFunc = lambda i, t: None,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        if len(editor_kwargs) < 2:
            msg = f"Need to pass at least 2 'editor_kwargs'. Passed {len(editor_kwargs)}"
            raise ValueError(msg)

        self._editor_aspects = tuple(
            frozenset(mapping.keys())
            for group in map(_to_group_factory_mapping, editor_kwargs.values())
            for mapping in group.values()
        )

        for a, b in itertools.combinations(self._editor_aspects, 2):
            if intersec := a.intersection(b):
                msg = f"Pidget factory mappings needs to be disjoint. Found duplicates {intersec}"
                raise ValueError(msg)

        if save_load_buttons:
            buttons = create_json_save_load_buttons_from_type(
                self, config_type, extra_presenter=extra_presenter
            )
            assert buttons is not None
            min_top_padding = buttons.sizeHint().height() // 2
        else:
            buttons = None
            min_top_padding = 0

        self._editors = tuple(
            AttrsConfigEditor(
                title=title,
                factory_mapping=mapping,
                config_type=config_type,
                save_load_buttons=False,
                min_top_padding=min_top_padding if i == 0 else 0,
                extra_presenter=extra_presenter,
            )
            for i, (title, mapping) in enumerate(editor_kwargs.items())
        )

        for editor in self._editors:
            editor.sig_update.connect(lambda _: self.sig_update.emit(self.get_data()))

        editor_layout = QVBoxLayout()
        for editor in self._editors:
            editor_layout.addWidget(editor)
        editor_layout.setContentsMargins(0, 0, 0, 0)

        if buttons is None:
            self.setLayout(editor_layout)
        else:
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(
                Overlay(
                    base=LayoutWrapper(editor_layout),
                    overlays=[(buttons, Overlay.positioner_top_right)],
                )
            )
            self.setLayout(layout)

    def get_data(self) -> Optional[T]:
        (head_editor, *tail_editors) = self._editors
        (_, *tail_editor_aspects) = self._editor_aspects

        instance = head_editor.get_data()
        if instance is None:
            return None

        for editor, aspects in zip(tail_editors, tail_editor_aspects):
            partial = editor.get_data()
            if partial is None:
                return None

            changes = {asp: getattr(partial, asp) for asp in aspects}
            instance = attrs.evolve(instance, **changes)  # type: ignore[misc]

        return instance

    def set_data(self, data: Optional[T]) -> None:
        for editor in self._editors:
            editor.set_data(data)

    def set_pidget_enabled(self, aspect: str, enabled: bool) -> None:
        for editor in self._editors:
            editor.set_pidget_enabled(aspect, enabled)

    def handle_validation_results(self, results: list[ValidationResult]) -> list[ValidationResult]:
        not_handled = results
        for editor in self._editors:
            not_handled = editor.handle_validation_results(not_handled)
        return not_handled

    @property
    def is_ready(self) -> bool:
        return all([editor.is_ready for editor in self._editors])



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from ._detector import Detector, DetectorConfig, DetectorContext, DetectorMetadata, DetectorResult
from ._processors import Processor, ProcessorConfig, ProcessorExtraResult, ProcessorResult



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import typing as t

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_eq, attrs_ndarray_isclose
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121.algo import (
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ExtendedProcessorBase,
)


class AmplitudeMethod(AlgoParamEnum):
    COHERENT = "Coherent"
    NONCOHERENT = "Noncoherent"
    FFT_MAX = "FFT max"


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    amplitude_method: AmplitudeMethod = attrs.field(
        default=AmplitudeMethod.COHERENT, converter=AmplitudeMethod
    )

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        return []


@attrs.frozen(kw_only=True)
class SubsweepProcessorResult:
    frame: npt.NDArray[np.complex128] = attrs.field(eq=attrs_ndarray_eq)
    amplitudes: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    phases: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    distance_velocity_map: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)


EntryResult = t.List[SubsweepProcessorResult]
ProcessorResult = t.List[t.Dict[int, EntryResult]]


class Processor(ExtendedProcessorBase[ProcessorResult]):
    def __init__(
        self,
        *,
        session_config: a121.SessionConfig,
        processor_config: ProcessorConfig,
    ) -> None:
        self.processor_config = processor_config
        self.processor_config.validate(session_config)

        self.windows = utils.map_over_extended_structure(
            self._get_hanning_widow, session_config.groups
        )

    @staticmethod
    def _get_hanning_widow(sensor_config: a121.SensorConfig) -> npt.NDArray[np.float64]:
        spf = sensor_config.sweeps_per_frame
        window = np.hanning(spf)[:, None]
        return window / np.sum(window)  # type: ignore[no-any-return]

    def _process_entry(
        self, result_hanning_window: t.Tuple[a121.Result, npt.NDArray[np.float64]]
    ) -> EntryResult:
        (result, hanning_window) = result_hanning_window

        entry_result = []

        for subframe in result.subframes:
            z_ft = np.fft.fftshift(np.fft.fft(subframe * hanning_window, axis=0), axes=(0,))
            abs_z_ft = np.abs(z_ft)

            amplitude_method = self.processor_config.amplitude_method
            if amplitude_method == AmplitudeMethod.COHERENT:
                ampls = np.abs(subframe.mean(axis=0))
            elif amplitude_method == AmplitudeMethod.NONCOHERENT:
                ampls = np.abs(subframe).mean(axis=0)
            elif amplitude_method == AmplitudeMethod.FFT_MAX:
                ampls = abs_z_ft.max(axis=0)
            else:
                msg = f"Unknown AmplitudeMethod: {amplitude_method}"
                raise RuntimeError(msg)

            phases = np.angle(subframe.mean(axis=0))

            entry_result.append(
                SubsweepProcessorResult(
                    frame=subframe, amplitudes=ampls, phases=phases, distance_velocity_map=abs_z_ft
                )
            )

        return entry_result

    def process(self, results: list[dict[int, a121.Result]]) -> ProcessorResult:
        return utils.map_over_extended_structure(
            self._process_entry, utils.zip_extended_structures(results, self.windows)
        )


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=32,
        num_points=40,
        step_length=8,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .setup_step import SetupStep


class WithDescription(SetupStep):
    """Decorates a SetupStep by `report`ing a description
    before reporting the step
    """

    def __init__(self, description: str, step: SetupStep) -> None:
        self.description = description
        self.step = step

    def report(self) -> None:
        print(self.description)
        self.step.report()
        print()

    def run(self) -> bool:
        return self.step.run()



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import copy
import logging
from functools import partial
from typing import Any, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool._core.entities.validation_result import Criticality
from acconeer.exptool.app.new.ui import icons
from acconeer.exptool.app.new.ui.components import pidgets
from acconeer.exptool.app.new.ui.components.a121 import RangeHelpView
from acconeer.exptool.app.new.ui.components.collapsible_widget import CollapsibleWidget
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor
from acconeer.exptool.app.new.ui.components.types import PidgetFactoryMapping


log = logging.getLogger(__name__)


class SubsweepConfigEditor(DataEditor[Optional[a121.SubsweepConfig]]):
    sig_update = Signal(object)

    _subsweep_config: Optional[a121.SubsweepConfig]

    _subsweep_config_pidgets: dict[str, pidgets.Pidget]
    _erroneous_aspects: set[str]

    SPACING = 15
    PROFILE_LABEL_MAP = {
        a121.Profile.PROFILE_1: "1 (shortest)",
        a121.Profile.PROFILE_2: "2",
        a121.Profile.PROFILE_3: "3",
        a121.Profile.PROFILE_4: "4",
        a121.Profile.PROFILE_5: "5 (longest)",
    }
    PRF_LABEL_MAP = {
        a121.PRF.PRF_19_5_MHz: "19.5 MHz",
        a121.PRF.PRF_15_6_MHz: "15.6 MHz",
        a121.PRF.PRF_13_0_MHz: "13.0 MHz",
        a121.PRF.PRF_8_7_MHz: "8.7 MHz",
        a121.PRF.PRF_6_5_MHz: "6.5 MHz",
        a121.PRF.PRF_5_2_MHz: "5.2 MHz",
    }
    SUBSWEEP_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "start_point": pidgets.IntPidgetFactory(
            name_label_text="Start point:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "start_point"),
        ),
        "num_points": pidgets.IntPidgetFactory(
            name_label_text="Number of points:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "num_points"),
            limits=(1, 4095),
        ),
        "step_length": pidgets.IntPidgetFactory(
            name_label_text="Step length:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "step_length"),
            limits=(1, None),
        ),
        "hwaas": pidgets.IntPidgetFactory(
            name_label_text="HWAAS:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "hwaas"),
            limits=(1, 511),
        ),
        "receiver_gain": pidgets.IntPidgetFactory(
            name_label_text="Receiver gain:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "receiver_gain"),
            limits=(0, 23),
        ),
        "profile": pidgets.EnumPidgetFactory(
            enum_type=a121.Profile,
            name_label_text="Profile:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "profile"),
            label_mapping=PROFILE_LABEL_MAP,
        ),
        "prf": pidgets.EnumPidgetFactory(
            enum_type=a121.PRF,
            name_label_text="PRF:",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "prf"),
            label_mapping=PRF_LABEL_MAP,
        ),
        "enable_tx": pidgets.CheckboxPidgetFactory(
            name_label_text="Enable transmitter",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "enable_tx"),
        ),
        "enable_loopback": pidgets.CheckboxPidgetFactory(
            name_label_text="Enable loopback",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "enable_loopback"),
        ),
        "phase_enhancement": pidgets.CheckboxPidgetFactory(
            name_label_text="Phase enhancement",
            name_label_tooltip=get_attribute_docstring(a121.SubsweepConfig, "phase_enhancement"),
        ),
        "iq_imbalance_compensation": pidgets.CheckboxPidgetFactory(
            name_label_text="IQ imbalance compensation",
            name_label_tooltip=get_attribute_docstring(
                a121.SubsweepConfig, "iq_imbalance_compensation"
            ),
        ),
    }
    ADVANCED_PARAMETERS = {
        "receiver_gain",
        "prf",
        "enable_tx",
        "enable_loopback",
        "phase_enhancement",
        "iq_imbalance_compensation",
    }

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)

        self._subsweep_config_pidgets = {}
        self._erroneous_aspects = set()

        self._subsweep_config = None

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.range_help_view = RangeHelpView(self)
        layout.addWidget(self.range_help_view)

        collapsible_layout = QVBoxLayout()

        for aspect, factory in self.SUBSWEEP_CONFIG_FACTORIES.items():
            pidget = factory.create(self)

            if aspect in self.ADVANCED_PARAMETERS:
                collapsible_layout.addWidget(pidget)
            else:
                layout.addWidget(pidget)

            pidget.sig_update.connect(partial(self._update_subsweep_config_aspect, aspect))

            self._subsweep_config_pidgets[aspect] = pidget

        # Some left margin here will show the hierarchy
        collapsible_layout.setContentsMargins(11, 0, 0, 0)
        dummy = QWidget()
        dummy.setLayout(collapsible_layout)
        self.collapsible_widget = CollapsibleWidget(label="Advanced", widget=dummy)

        layout.addWidget(self.collapsible_widget)

        self.setLayout(layout)

    def set_read_only(self, read_only: bool) -> None:
        for pidget in self._subsweep_config_pidgets.values():
            pidget.setEnabled(not read_only)

    def set_data(self, subsweep_config: Optional[a121.SubsweepConfig]) -> None:
        if self._subsweep_config == subsweep_config:
            return

        self.range_help_view.set_data(subsweep_config)
        self._subsweep_config = subsweep_config

        self.setEnabled(self._subsweep_config is not None)
        if self._subsweep_config is None:
            log.debug("could not update ui as SubsweepConfig is None")
            return

        for aspect, pidget in self._subsweep_config_pidgets.items():
            if aspect in self._erroneous_aspects:
                continue
            pidget.set_data(getattr(self._subsweep_config, aspect))

    def get_data(self) -> Optional[a121.SubsweepConfig]:
        return copy.deepcopy(self._subsweep_config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        current_data = self.get_data()
        advanced_results = [
            r
            for r in results
            if (r.aspect in self.ADVANCED_PARAMETERS) and (r.source == current_data)
        ]

        if any(result.criticality is Criticality.ERROR for result in advanced_results):
            self.collapsible_widget.set_icon(icons.WARNING(color=icons.ERROR_RED))
        elif any(result.criticality is Criticality.WARNING for result in advanced_results):
            self.collapsible_widget.set_icon(icons.WARNING())
        else:
            self.collapsible_widget.set_icon(None)

        for aspect, pidget in self._subsweep_config_pidgets.items():
            if aspect not in self._erroneous_aspects:
                pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return self._erroneous_aspects == set()

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None or self._subsweep_config is None:
            return False

        if result.aspect in self._erroneous_aspects:
            # If there is an erroneous aspect in the GUI, we do not want to overwrite that.
            # Once the erroneous aspect is handled with, the same validation result will
            # come through here anyway
            return True

        result_handled = False

        if result.source == self._subsweep_config:
            self._subsweep_config_pidgets[result.aspect].set_note_text(
                result.message, result.criticality
            )
            result_handled = True

        return result_handled

    def _update_subsweep_config_aspect(self, aspect: str, value: Any) -> None:
        if self._subsweep_config is None:
            msg = "SubsweepConfig is None"
            raise TypeError(msg)

        config = copy.deepcopy(self._subsweep_config)

        try:
            setattr(config, aspect, value)
        except Exception as e:
            self._erroneous_aspects.add(aspect)
            self._subsweep_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._erroneous_aspects.discard(aspect)
            self.set_data(config)
            self.sig_update.emit(config)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from ._bin_fetcher import (
    BIN_FETCH_PROMPT,
    ET_DIR,
    clear_cookies,
    download,
    get_content,
    get_cookies,
    login,
    save_cookies,
)
from ._dev_license import DevLicense
from ._flash_exception import FlashException
from ._flasher import (
    flash_image,
    get_boot_description,
    get_flash_download_name,
    get_flash_known_devices,
    get_post_dfu_description,
    is_fw_downloadable,
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import collections
import typing as t
from enum import Enum, auto

import attrs
import typing_extensions as te

from acconeer.exptool import a121


_ms = _mA = 1e-3
_us = _uA = 1e-6


@attrs.frozen
class Sensor:
    """
    Representation of a sensor.

    Consists of multiple table for different currents, overhead- and transition times.
    """

    class IdleState(Enum):
        OFF = auto()
        HIBERNATE = auto()

    class FixedOverhead(Enum):
        SUBSWEEP = auto()
        SWEEP = auto()
        FRAME = auto()

    LowerIdleState = te.Literal[IdleState.OFF, IdleState.HIBERNATE]

    inter_sweep_currents: t.Mapping[a121.IdleState, float]
    inter_sweep_currents_high_speed_mode: t.Mapping[a121.IdleState, float]
    inter_frame_currents: t.Mapping[t.Union[a121.IdleState, IdleState], float]
    measure_currents: t.Mapping[a121.Profile, float]

    time_for_measure_transition: t.Mapping[a121.IdleState, float]
    normal_overheads: t.Mapping[FixedOverhead, float]
    high_speed_mode_overheads: t.Mapping[FixedOverhead, float]
    pre_measure_setup_duration: float
    read_duration: float

    sample_durations: t.Mapping[a121.Profile, t.Mapping[a121.PRF, float]]
    point_overheads: t.Mapping[a121.Profile, t.Mapping[a121.PRF, float]]

    @classmethod
    def none(cls) -> te.Self:
        return attrs.evolve(
            cls.default(),
            inter_sweep_currents=collections.defaultdict(lambda: 0.0),
            inter_sweep_currents_high_speed_mode=collections.defaultdict(lambda: 0.0),
            inter_frame_currents=collections.defaultdict(lambda: 0.0),
            measure_currents=collections.defaultdict(lambda: 0.0),
        )

    @classmethod
    def default(cls) -> te.Self:
        idle_currents = {
            a121.IdleState.READY: 62.5 * _mA,
            a121.IdleState.SLEEP: 1.6 * _mA,
            a121.IdleState.DEEP_SLEEP: 1.1 * _mA,
        }

        return cls(
            inter_sweep_currents_high_speed_mode={
                a121.IdleState.READY: 68.5 * _mA,
                a121.IdleState.SLEEP: 1.6 * _mA,
                a121.IdleState.DEEP_SLEEP: 1.1 * _mA,
            },
            inter_sweep_currents=idle_currents,
            inter_frame_currents={
                **idle_currents,  # type: ignore[dict-item]
                cls.IdleState.OFF: 3 * _uA,
                cls.IdleState.HIBERNATE: 15 * _uA,
            },
            measure_currents={
                a121.Profile.PROFILE_1: 73.0 * _mA,
                a121.Profile.PROFILE_2: 72.5 * _mA,
                a121.Profile.PROFILE_3: 74.0 * _mA,
                a121.Profile.PROFILE_4: 75.0 * _mA,
                a121.Profile.PROFILE_5: 76.5 * _mA,
            },
            time_for_measure_transition={
                a121.IdleState.READY: 0,
                a121.IdleState.SLEEP: 55 * _us,
                a121.IdleState.DEEP_SLEEP: 670 * _us,
            },
            normal_overheads={
                cls.FixedOverhead.SUBSWEEP: 22 * _us,
                cls.FixedOverhead.SWEEP: 10 * _us,
                cls.FixedOverhead.FRAME: 4 * _us,
            },
            high_speed_mode_overheads={
                cls.FixedOverhead.SUBSWEEP: 0,
                cls.FixedOverhead.SWEEP: 0,
                cls.FixedOverhead.FRAME: 36 * _us,
            },
            pre_measure_setup_duration=600 * _us,
            read_duration=0.2 * _ms,
            sample_durations={
                a121.Profile.PROFILE_1: {
                    a121.PRF.PRF_19_5_MHz: 1.487 * _us,
                    a121.PRF.PRF_15_6_MHz: 1.795 * _us,
                    a121.PRF.PRF_13_0_MHz: 2.103 * _us,
                    a121.PRF.PRF_8_7_MHz: 3.026 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.949 * _us,
                    a121.PRF.PRF_5_2_MHz: 4.872 * _us,
                },
                a121.Profile.PROFILE_2: {
                    a121.PRF.PRF_15_6_MHz: 1.344 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.600 * _us,
                    a121.PRF.PRF_8_7_MHz: 2.369 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.138 * _us,
                    a121.PRF.PRF_5_2_MHz: 3.908 * _us,
                },
                a121.Profile.PROFILE_3: {
                    a121.PRF.PRF_15_6_MHz: 1.026 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.231 * _us,
                    a121.PRF.PRF_8_7_MHz: 1.846 * _us,
                    a121.PRF.PRF_6_5_MHz: 2.462 * _us,
                    a121.PRF.PRF_5_2_MHz: 3.077 * _us,
                },
                a121.Profile.PROFILE_4: {
                    a121.PRF.PRF_15_6_MHz: 1.026 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.231 * _us,
                    a121.PRF.PRF_8_7_MHz: 1.846 * _us,
                    a121.PRF.PRF_6_5_MHz: 2.462 * _us,
                    a121.PRF.PRF_5_2_MHz: 3.077 * _us,
                },
                a121.Profile.PROFILE_5: {
                    a121.PRF.PRF_15_6_MHz: 1.026 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.231 * _us,
                    a121.PRF.PRF_8_7_MHz: 1.846 * _us,
                    a121.PRF.PRF_6_5_MHz: 2.462 * _us,
                    a121.PRF.PRF_5_2_MHz: 3.077 * _us,
                },
            },
            point_overheads={
                a121.Profile.PROFILE_1: {
                    a121.PRF.PRF_19_5_MHz: 1.744 * _us,
                    a121.PRF.PRF_15_6_MHz: 2.102 * _us,
                    a121.PRF.PRF_13_0_MHz: 2.462 * _us,
                    a121.PRF.PRF_8_7_MHz: 3.539 * _us,
                    a121.PRF.PRF_6_5_MHz: 4.615 * _us,
                    a121.PRF.PRF_5_2_MHz: 5.692 * _us,
                },
                a121.Profile.PROFILE_2: {
                    a121.PRF.PRF_15_6_MHz: 1.612 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.920 * _us,
                    a121.PRF.PRF_8_7_MHz: 2.844 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.766 * _us,
                    a121.PRF.PRF_5_2_MHz: 4.689 * _us,
                },
                a121.Profile.PROFILE_3: {
                    a121.PRF.PRF_15_6_MHz: 1.282 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.539 * _us,
                    a121.PRF.PRF_8_7_MHz: 2.308 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.077 * _us,
                    a121.PRF.PRF_5_2_MHz: 4.689 * _us,
                },
                a121.Profile.PROFILE_4: {
                    a121.PRF.PRF_15_6_MHz: 1.282 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.539 * _us,
                    a121.PRF.PRF_8_7_MHz: 2.308 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.077 * _us,
                    a121.PRF.PRF_5_2_MHz: 4.689 * _us,
                },
                a121.Profile.PROFILE_5: {
                    a121.PRF.PRF_15_6_MHz: 1.282 * _us,
                    a121.PRF.PRF_13_0_MHz: 1.539 * _us,
                    a121.PRF.PRF_8_7_MHz: 2.308 * _us,
                    a121.PRF.PRF_6_5_MHz: 3.077 * _us,
                    a121.PRF.PRF_5_2_MHz: 4.689 * _us,
                },
            },
        )


@attrs.frozen
class Module:
    """
    Representation of a module.

    Consists of a current table for the different module power states.
    """

    class PowerState(Enum):
        PROCESSING = auto()
        AWAKE = auto()
        ASLEEP = auto()
        IO = auto()

    currents: t.Mapping[t.Union[PowerState, Sensor.IdleState], float]

    @classmethod
    def xm125(cls) -> te.Self:
        return cls(
            currents={
                cls.PowerState.PROCESSING: 13 * _mA,
                cls.PowerState.AWAKE: 4.8 * _mA,
                cls.PowerState.ASLEEP: 0.1 * _uA,
                cls.PowerState.IO: 13 * _mA,
            }
        )

    @classmethod
    def none(cls) -> te.Self:
        return cls(currents=collections.defaultdict(lambda: 0.0))
